from pysb.export.stochkit import StochKitExporter
from pysb.simulator import StochKitSimulator
from pysb.examples import robertson, earm_1_0, expression_observables
import numpy as np
from pysb.core import as_complex_pattern

_STOCHKIT_SEED = 123


def test_stochkit_export():
    StochKitExporter(robertson.model).export()


def test_stochkit_earm():
    tspan = np.linspace(0, 1000, 10)
    sim = StochKitSimulator(earm_1_0.model, tspan=tspan)
    simres = sim.run(n_runs=2, seed=_STOCHKIT_SEED, algorithm="ssa")
    simres_tl = sim.run(n_runs=2, seed=_STOCHKIT_SEED, algorithm="tau_leaping")


def test_stochkit_earm_multi_initials():
    model = earm_1_0.model
    tspan = np.linspace(0, 1000, 10)
    sim = StochKitSimulator(model, tspan=tspan)
    unbound_L = model.monomers['L'](b=None)
    simres = sim.run(initials={unbound_L: [3000, 1500]},
                     n_runs=2, seed=_STOCHKIT_SEED, algorithm="ssa")
    df = simres.dataframe

    unbound_L_index = model.get_species_index(as_complex_pattern(unbound_L))

    # Check we have two repeats of each initial
    assert np.allclose(df.loc[(slice(None), 0), '__s%d' % unbound_L_index],
                       [3000, 3000, 1500, 1500])


def test_stochkit_expressions():
    model = expression_observables.model
    tspan = np.linspace(0, 100, 11)
    sim = StochKitSimulator(model, tspan=tspan)
    assert np.allclose(sim.run().tout, tspan)



#!/usr/bin/env python
"""Simulate the bax_pore model and plot the results."""

from __future__ import print_function
import matplotlib.pyplot as plt
from numpy import linspace
from pysb.simulator import ScipyOdeSimulator

from bax_pore import model


t = linspace(0, 100)
print("Simulating...")
x = ScipyOdeSimulator(model).run(tspan=t).all

plt.plot(t, x['BAX4'])
plt.plot(t, x['BAX4_inh'])
plt.legend(['BAX4', 'BAX4_inh'], loc='upper left')
plt.show()



import collections
from .core import ComplexPattern, MonomerPattern, Monomer, \
    ReactionPattern, ANY, as_complex_pattern, DanglingBondError
import networkx as nx
from networkx.algorithms.isomorphism.vf2userfunc import GraphMatcher
from networkx.algorithms.isomorphism import categorical_node_match
import numpy as np
try:
    basestring
except NameError:
    # Under Python 3, do not pretend that bytes are a valid string
    basestring = str


def get_half_bonds_in_pattern(pat):
    """
    Return the list of integer bond numbers used in a pattern

    To return as a set, use :func:`get_bonds_in_pattern`.

    Parameters
    ----------
    pat : ComplexPattern, MonomerPattern, or None
        A pattern from which bond numberings are extracted

    Returns
    -------
    list
        Bond numbers used in the supplied pattern

    Examples
    --------

    >>> A = Monomer('A', ['b1', 'b2'], _export=False)
    >>> get_half_bonds_in_pattern(A(b1=None, b2=None))
    []
    >>> get_half_bonds_in_pattern(A(b1=1) % A(b2=1))
    [1, 1]
    """
    bonds_used = list()

    def _get_bonds_in_monomer_pattern(mp):
        for sc in mp.site_conditions.values():
            if isinstance(sc, int):
                bonds_used.append(sc)
            elif not isinstance(sc, basestring) and \
                    isinstance(sc, collections.Iterable):
                [bonds_used.append(b) for b in sc if isinstance(b, int)]

    if pat is None:
        return bonds_used
    if isinstance(pat, MonomerPattern):
        _get_bonds_in_monomer_pattern(pat)
    elif isinstance(pat, ComplexPattern):
        for mp in pat.monomer_patterns:
            _get_bonds_in_monomer_pattern(mp)
    else:
        raise ValueError('Unknown pattern type: %s' % type(pat))

    return bonds_used


def get_bonds_in_pattern(pat):
    """
    Return the set of integer bond numbers used in a pattern

    To return as a list (with duplicates), use
    :func:`get_half_bonds_in_pattern`

    Parameters
    ----------
    pat : ComplexPattern, MonomerPattern, or None
        A pattern from which bond numberings are extracted

    Returns
    -------
    set
        Bond numbers used in the supplied pattern

    Examples
    --------

    >>> A = Monomer('A', ['b1', 'b2'], _export=False)
    >>> get_bonds_in_pattern(A(b1=None, b2=None)) == set()
    True
    >>> get_bonds_in_pattern(A(b1=1) % A(b2=1)) == {1}
    True
    >>> get_bonds_in_pattern(A(b1=1) % A(b1=2, b2=1) % A(b1=2)) == {1, 2}
    True
    """
    return set(get_half_bonds_in_pattern(pat))


def check_dangling_bonds(pattern):
    """
    Check for dangling bonds in a PySB ComplexPattern/ReactionPattern

    Raises a DanglingBondError if a dangling bond is found
    """
    if isinstance(pattern, ReactionPattern):
        for cp in pattern.complex_patterns:
            check_dangling_bonds(cp)
        return
    bond_counts = collections.Counter(get_half_bonds_in_pattern(pattern))

    dangling_bonds = [bond for bond, count in bond_counts.items()
                      if count == 1]
    if dangling_bonds:
        raise DanglingBondError('Dangling bond(s) {} in {}'
                                .format(dangling_bonds, pattern))


def _match_graphs(pattern, candidate, exact):
    """ Compare two pattern graphs for isomorphism """
    node_matcher = categorical_node_match('id', default=None)
    if exact:
        return nx.is_isomorphic(pattern._as_graph(),
                                candidate._as_graph(),
                                node_match=node_matcher)
    else:
        return GraphMatcher(
            candidate._as_graph(), pattern._as_graph(),
            node_match=node_matcher
        ).subgraph_is_isomorphic()


def match_complex_pattern(pattern, candidate, exact=False):
    """
    Compare two ComplexPatterns against each other

    Parameters
    ----------
    pattern: pysb.ComplexPattern
    candidate: pysb.Complex.Pattern
    exact: bool
        Set to True for exact matches (i.e. species equivalence,
        or exact graph isomorphism). Set to False to compare as a
        pattern (i.e. subgraph isomorphism).

    Returns
    -------
    True if pattern matches candidate, False otherwise
    """
    if exact:
        if not pattern.is_concrete():
            raise ValueError('Pattern must be concrete for '
                             'exact matching: {}'.format(pattern))
        if not candidate.is_concrete():
            raise ValueError('Candidate must be concrete for '
                             'exact matching: {}'.format(candidate))

    if exact and len(pattern.monomer_patterns) != len(
            candidate.monomer_patterns):
        return False

    # Compare the monomer counts in the patterns so we can fail fast
    # without having to compare bonds using graph isomorphism checks, which
    # are more computationally expensive
    mons_pat = collections.Counter([mp.monomer for mp in \
            pattern.monomer_patterns])
    mons_cand = collections.Counter([mp.monomer for mp in \
            candidate.monomer_patterns])

    for mon, mon_count_cand in mons_cand.items():
        mon_count_pat = mons_pat.get(mon, 0)
        if exact and mon_count_cand != mon_count_pat:
            return False
        if mon_count_pat > mon_count_cand:
            return False

    # If we've got this far, we'll need to do a full pattern match
    # by searching for a graph isomorphism
    return _match_graphs(pattern, candidate, exact=exact)


def match_reaction_pattern(pattern, candidate):
    """
    Compare two ReactionPatterns against each other

    This function tests that every ComplexPattern in pattern has a
    matching ComplexPattern in candidate. If there's a one-to-one
    mapping of ComplexPattern matches, this is straightforward.
    Otherwise, we need to check for a maximum matching - a graph theory
    term referring to the maximum number of edges possible in a
    bipartite graph (representing ComplexPattern compatibility between
    pattern and candidate) without overlapping nodes. If every
    ComplexPattern in pattern has a match, then return True, otherwise
    return False. This algorithm is polynomial time (although the
    ComplexPattern isomorphism comparisons using match_complex_pattern are
    not).

    Parameters
    ----------
    pattern: pysb.ReactionPattern
    candidate: pysb.ReactionPattern

    Returns
    -------
    True if pattern matches candidate, False otherwise.

    """
    if len(pattern.complex_patterns) > len(candidate.complex_patterns):
        return False

    matches = []
    for cplx_pat in pattern.complex_patterns:
        matches_this = [cand_cplx_pat.matches(
            cplx_pat) for cand_cplx_pat in
            candidate.complex_patterns]
        matches_this = set(np.where(matches_this)[0])
        if len(matches_this) == 0:
            return False
        matches.append(matches_this)

    # If a unique 1:1 mapping exists, a match is assured
    if len(set.intersection(*matches)) == 0:
        return True

    # Find the maximum matching in a bipartite graph representing the
    # two sets of ComplexPatterns
    g = nx.Graph()
    g.add_nodes_from(['p%d' % n for n in
                     range(len(pattern.complex_patterns))], bipartite=0)
    g.add_nodes_from(['c%d' % n for n in
                     range(len(candidate.complex_patterns))], bipartite=1)
    for src_pat_id, src_pat_matches in enumerate(matches):
        g.add_edges_from([('p%d' % src_pat_id, 'c%d' % cand_pat_id) for
                          cand_pat_id in src_pat_matches])

    return (len(nx.bipartite.maximum_matching(g)) // 2) == len(
        pattern.complex_patterns)


def monomers_from_pattern(pattern):
    """ Return the set of monomers used in a pattern """
    if isinstance(pattern, ReactionPattern):
        return set.union(*[monomers_from_pattern(cp)
                           for cp in pattern.complex_patterns])
    if isinstance(pattern, ComplexPattern):
        return set([mp.monomer for mp in pattern.monomer_patterns])
    elif isinstance(pattern, MonomerPattern):
        return {pattern.monomer}
    elif isinstance(pattern, Monomer):
        return {pattern}
    else:
        raise Exception('Unsupported pattern type: %s' % type(pattern))


class SpeciesPatternMatcher(object):
    """
    Match a pattern against a model's species list

    Examples
    --------

    Create a PatternMatcher for the EARM 1.0 model

    >>> from pysb.examples.earm_1_0 import model
    >>> from pysb.bng import generate_equations
    >>> from pysb.pattern import SpeciesPatternMatcher
    >>> from pysb import ANY, WILD, Model, Monomer, as_complex_pattern
    >>> generate_equations(model)
    >>> spm = SpeciesPatternMatcher(model)

    Assign two monomers to variables (only needed when importing a model
    instead of defining one interactively)

    >>> Bax4 = model.monomers['Bax4']
    >>> Bcl2 = model.monomers['Bcl2']

    Search using a Monomer

    >>> spm.match(Bax4)
    [Bax4(b=None), Bax4(b=1) % Bcl2(b=1), Bax4(b=1) % Mito(b=1)]
    >>> spm.match(Bcl2) # doctest:+NORMALIZE_WHITESPACE
    [Bax2(b=1) % Bcl2(b=1),
    Bax4(b=1) % Bcl2(b=1),
    Bcl2(b=None),
    Bcl2(b=1) % MBax(b=1)]

    Search using a MonomerPattern (ANY and WILD keywords can be used)

    >>> spm.match(Bax4(b=WILD))
    [Bax4(b=None), Bax4(b=1) % Bcl2(b=1), Bax4(b=1) % Mito(b=1)]
    >>> spm.match(Bcl2(b=ANY))
    [Bax2(b=1) % Bcl2(b=1), Bax4(b=1) % Bcl2(b=1), Bcl2(b=1) % MBax(b=1)]

    Search using a ComplexPattern

    >>> spm.match(Bax4(b=1) % Bcl2(b=1))
    [Bax4(b=1) % Bcl2(b=1)]
    >>> spm.match(Bax4() % Bcl2())
    [Bax4(b=1) % Bcl2(b=1)]

    Contrived example to test a site with both a bond and state defined

    >>> model = Model(_export=False)
    >>> A = Monomer('A', ['a'], {'a': ['u', 'p']}, _export=False)
    >>> model.add_component(A)
    >>> species = [                                                     \
            A(a=None),                                                  \
            A(a='u'),                                                   \
            A(a=1) % A(a=1),                                            \
            A(a=('u', 1)) % A(a=('u', 1)),                              \
            A(a=('p', 1)) % A(a=('p', 1))                               \
        ]
    >>> model.species = [as_complex_pattern(sp) for sp in species]
    >>> spm2 = SpeciesPatternMatcher(model)
    >>> spm2.match(A()) # doctest:+NORMALIZE_WHITESPACE
    [A(a=None), A(a='u'), A(a=1) % A(a=1), A(a=('u', 1)) % A(a=('u', 1)),
     A(a=('p', 1)) % A(a=('p', 1))]
    >>> spm2.match(A(a='u'))
    [A(a='u')]
    >>> spm2.match(A(a=('u', ANY)))
    [A(a=('u', 1)) % A(a=('u', 1))]
    >>> spm2.match(A(a=('u', WILD)))
    [A(a='u'), A(a=('u', 1)) % A(a=('u', 1))]
    """
    def __init__(self, model, species=None):
        self.model = model
        if not species and not model.species:
            raise Exception('Model needs species list - run '
                            'generate_equations() first')

        if not species:
            species = model.species

        self.species = species

        self._species_cache = collections.defaultdict(set)
        for idx, sp in enumerate(species):
            self._add_species(idx, sp)

    def _add_species(self, idx, sp):
        if sp.compartment:
            raise NotImplementedError
        for mp in sp.monomer_patterns:
            if mp.compartment:
                raise NotImplementedError
            self._species_cache[mp.monomer].add(idx)

    def add_species(self, species, check_duplicate=True):
        """
        Add a species to the search list without adding to the model

        Parameters
        ----------
        species : ComplexPattern
            A concrete ComplexPattern (molecular species) to add to the
            search list
        check_duplicate : bool, optional
            If True, check the species list to make sure the new species
            is not already in the list
        """
        if check_duplicate and self.match(species, exact=True):
            return
        self.species.append(species)
        self._add_species(len(self.species) - 1, species)

    def match(self, pattern, index=False, exact=False):
        """
        Match a pattern against the list of species

        Parameters
        ----------
        pattern: pysb.Monomer or pysb.MonomerPattern or pysb.ComplexPattern
        index: bool
            If True, return species numerical index, rather than species itself
        exact: bool
            Treat Match as exact equivalence, not a pattern match (i.e. must be
            concrete if a MonomerPattern or ComplexPattern)

        Returns
        -------
        list of pysb.ComplexPattern or list of int
            A list of species matching the pattern is returned, unless
            index=True, in which case a list of the numerical indices of
            matching species is returned instead

        Examples
        --------

        >>> from pysb.examples import earm_1_0
        >>> from pysb.bng import generate_equations
        >>> model = earm_1_0.model
        >>> generate_equations(model)
        >>> spm = SpeciesPatternMatcher(model)
        >>> L = model.monomers['L']
        >>> spm.match(L())
        [L(b=None), L(b=1) % pR(b=1)]
        """
        if not isinstance(pattern, (Monomer, MonomerPattern, ComplexPattern)):
            raise ValueError('A Monomer, MonomerPattern or ComplexPattern is '
                             'required to match species')

        monomers = monomers_from_pattern(pattern)

        if exact:
            if isinstance(pattern, (Monomer, MonomerPattern)):
                num_mon_pats = 1
            else:
                num_mon_pats = len(pattern.monomer_patterns)
        else:
            # Don't check the number of monomer patterns in search
            # candidates if we're not doing an exact match of the species
            num_mon_pats = None

        shortlist, shortlist_indexes = self._species_containing_monomers(
            monomers, num_mon_pats)

        # If pattern is a Monomer, we're done
        if isinstance(pattern, Monomer):
            return shortlist_indexes if index else shortlist
        else:
            return [(shortlist_indexes[idx] if index else sp) for idx, sp in
                    enumerate(shortlist) if
                    match_complex_pattern(
                        as_complex_pattern(pattern), sp, exact=exact
                    )]

    def _species_containing_monomers(self, monomer_list, num_mon_pats=None):
        """
        Identifies species containing a list of monomers

        Parameters
        ----------
        monomer_list: list of Monomers
            A list of monomers with which to search the model's species
        num_mon_pats: int or None
            Restrict matches to species with exactly the specified number of
            MonomerPatterns

        Returns
        -------
        Model species containing all of the monomers in the list
        """
        sp_indexes = set.intersection(*[self._species_cache[mon] for mon in
                                        monomer_list])
        if num_mon_pats:
            retval = zip(*[(self.species[sp], sp) for sp in sp_indexes
                           if len(self.species[sp].monomer_patterns)
                           == num_mon_pats])
            return retval if retval else ((), ())
        else:
            return [self.species[sp] for sp in sp_indexes], list(sp_indexes)

    def rule_firing_species(self, rules_to_consider=None,
                            include_reverse=True):
        """
        Return the species which match the reactants of a set of rules

        Parameters
        ----------
        rules_to_consider: list of pysb.Rule or None
            A list of rules to use. If None, use all rules in the model.
        include_reverse: bool, optional
            For reversible rules, include species triggering the rule in
            reverse

        Returns
        -------
        collections.OrderedDict
            Dictionary of PySB rules whose reactants contain at least one of
            the species in the model. Keys are PySB rules, values are a list
            of lists. Each outer list corresponding to each
            ComplexPattern in the ReactantPattern (or ReactantPattern and
            ProductPattern, if rule is reversible). Each inner list contains
            the list of species matching the corresponding ComplexPattern.

        Examples
        --------

        >>> from pysb.examples import robertson
        >>> from pysb.bng import generate_equations
        >>> model = robertson.model
        >>> generate_equations(model)
        >>> spm = SpeciesPatternMatcher(model)

        Get a list of species which fire each rule:

        >>> spm.rule_firing_species() \
                #doctest: +NORMALIZE_WHITESPACE
        OrderedDict([(Rule('A_to_B', A() >> B(), k1), [[A()]]),
         (Rule('BB_to_BC', B() + B() >> B() + C(), k2), [[B()], [B()]]),
         (Rule('BC_to_AC', B() + C() >> A() + C(), k3), [[B()], [C()]])])
        """
        if rules_to_consider is None:
            rules_to_consider = self.model.rules
        rules_fired = collections.OrderedDict()
        for r in rules_to_consider:
            rp = r.reactant_pattern
            if len(rp.complex_patterns) == 0:
                # Synthesis rules are always fired
                rules_fired[r] = []
            else:
                species_fired = self.species_fired_by_reactant_pattern(rp)
                if include_reverse and r.is_reversible:
                    species_fired += self.species_fired_by_reactant_pattern(
                        r.product_pattern)
                if species_fired:
                    rules_fired[r] = species_fired
        return rules_fired

    def species_fired_by_reactant_pattern(self, reaction_pattern):
        """
        Get list of species matching a reactant pattern

        Parameters
        ----------
        reaction_pattern: pysb.ReactionPattern

        Returns
        -------
        list of lists of pysb.ComplexPattern
            List of lists of species matching each ComplexPattern in the
            ReactantPattern.

        Examples
        --------

        >>> from pysb.examples import bax_pore
        >>> from pysb.bng import generate_equations
        >>> model = bax_pore.model
        >>> generate_equations(model)
        >>> spm = SpeciesPatternMatcher(model)

        Get a list of species which fire each rule:

        >>> rxn_pat = model.rules['bax_dim'].reactant_pattern
        >>> print(rxn_pat)
        BAX(t1=None, t2=None) + BAX(t1=None, t2=None)

        >>> spm.species_fired_by_reactant_pattern(rxn_pat) \
                #doctest: +NORMALIZE_WHITESPACE
        [[BAX(t1=None, t2=None, inh=None),
          BAX(t1=None, t2=None, inh=1) % MCL1(b=1)],
         [BAX(t1=None, t2=None, inh=None),
              BAX(t1=None, t2=None, inh=1) % MCL1(b=1)]]
        """
        species_fired = []

        for i, cp in enumerate(reaction_pattern.complex_patterns):
            species_fired_this_cp = self.match(cp)
            if not species_fired_this_cp:
                return []
            else:
                species_fired.append(species_fired_this_cp)

        return species_fired


class RulePatternMatcher(object):
    """
    Match a pattern against a model's species list

    Methods are provided to match against rule reactants, products or both.
    Searches can be Monomers, MonomerPatterns, ComplexPatterns or
    ReactionPatterns.

    Examples
    --------

    Create a PatternMatcher for the EARM 1.0 model

    >>> from pysb.examples.earm_1_0 import model
    >>> from pysb.pattern import RulePatternMatcher
    >>> rpm = RulePatternMatcher(model)

    Assign some monomers to variables (only needed when importing a model
    instead of defining one interactively)

    >>> AMito, mCytoC, mSmac, cSmac = [model.monomers[m] for m in \
        ('AMito', 'mCytoC', 'mSmac', 'cSmac')]

    Search using a Monomer

    >>> rpm.match_reactants(AMito) # doctest:+NORMALIZE_WHITESPACE
    [Rule('bind_mCytoC_AMito', AMito(b=None) + mCytoC(b=None) |
        AMito(b=1) % mCytoC(b=1), kf20, kr20),
    Rule('produce_ACytoC_via_AMito', AMito(b=1) % mCytoC(b=1) >>
        AMito(b=None) + ACytoC(b=None), kc20),
    Rule('bind_mSmac_AMito', AMito(b=None) + mSmac(b=None) |
        AMito(b=1) % mSmac(b=1), kf21, kr21),
    Rule('produce_ASmac_via_AMito', AMito(b=1) % mSmac(b=1) >>
        AMito(b=None) + ASmac(b=None), kc21)]

    >>> rpm.match_products(mSmac) # doctest:+NORMALIZE_WHITESPACE
    [Rule('bind_mSmac_AMito', AMito(b=None) + mSmac(b=None) |
        AMito(b=1) % mSmac(b=1), kf21, kr21)]

    Search using a MonomerPattern

    >>> rpm.match_reactants(AMito(b=1)) # doctest:+NORMALIZE_WHITESPACE
    [Rule('produce_ACytoC_via_AMito', AMito(b=1) % mCytoC(b=1) >>
        AMito(b=None) + ACytoC(b=None), kc20),
    Rule('produce_ASmac_via_AMito', AMito(b=1) % mSmac(b=1) >>
        AMito(b=None) + ASmac(b=None), kc21)]

    >>> rpm.match_rules(cSmac(b=1)) # doctest:+NORMALIZE_WHITESPACE
    [Rule('inhibit_cSmac_by_XIAP', cSmac(b=None) + XIAP(b=None) |
        cSmac(b=1) % XIAP(b=1), kf28, kr28)]

    Search using a ComplexPattern

    >>> rpm.match_reactants(AMito() % mSmac()) # doctest:+NORMALIZE_WHITESPACE
    [Rule('produce_ASmac_via_AMito', AMito(b=1) % mSmac(b=1) >>
        AMito(b=None) + ASmac(b=None), kc21)]

    >>> rpm.match_rules(AMito(b=1) % mCytoC(b=1)) \
        # doctest:+NORMALIZE_WHITESPACE
    [Rule('bind_mCytoC_AMito', AMito(b=None) + mCytoC(b=None) |
        AMito(b=1) % mCytoC(b=1), kf20, kr20),
    Rule('produce_ACytoC_via_AMito', AMito(b=1) % mCytoC(b=1) >>
        AMito(b=None) + ACytoC(b=None), kc20)]

    Search using a ReactionPattern

    >>> rpm.match_reactants(mCytoC() + mSmac())
    []

    >>> rpm.match_reactants(AMito() + mCytoC()) # doctest:+NORMALIZE_WHITESPACE
    [Rule('bind_mCytoC_AMito', AMito(b=None) + mCytoC(b=None) |
        AMito(b=1) % mCytoC(b=1), kf20, kr20)]

    """

    def __init__(self, model):
        self.model = model

        self._reactant_cache = collections.defaultdict(set)
        self._product_cache = collections.defaultdict(set)

        for rule in model.rules:
            for cache, rp in ((self._reactant_cache, rule.reactant_pattern),
                              (self._product_cache, rule.product_pattern)):
                for cp in rp.complex_patterns:
                    if cp.compartment:
                        raise NotImplementedError
                    for mp in cp.monomer_patterns:
                        if mp.compartment:
                            raise NotImplementedError
                        cache[mp.monomer].add(rule.name)

    def match_reactants(self, pattern):
        return self._match_reaction_patterns(pattern, 'reactant')

    def match_products(self, pattern):
        return self._match_reaction_patterns(pattern, 'product')

    def match_rules(self, pattern):
        return [r for r in self.model.rules if
                r in self.match_reactants(pattern) or
                r in self.match_products(pattern)]

    def _match_reaction_patterns(self, pattern, reaction_side):
        if not isinstance(pattern, (Monomer, MonomerPattern, ComplexPattern,
                                    ReactionPattern)):
            raise ValueError('A Monomer, MonomerPattern, ComplexPattern or '
                             'ReactionPattern required to match rules')

        monomers = monomers_from_pattern(pattern)

        if reaction_side == 'reactant':
            cache = self._reactant_cache

            def pat_fn(r):
                return r.reactant_pattern
        elif reaction_side == 'product':
            cache = self._product_cache

            def pat_fn(r):
                return r.product_pattern
        else:
            raise Exception('reaction_side must be "reactant" or "product"')

        shortlist = self._cache_containing_monomers(cache, monomers)

        # If pattern is a Monomer, we're done
        if isinstance(pattern, Monomer):
            return shortlist

        if isinstance(pattern, (MonomerPattern, ComplexPattern)):
            new_shortlist = []
            for rule in shortlist:
                reaction_pattern = pat_fn(rule)
                if self._match_complex_pattern_to_reaction_pattern(
                        as_complex_pattern(pattern), reaction_pattern):
                    new_shortlist.append(rule)

            return new_shortlist

        else:
            return [rule for rule in shortlist if
                    pat_fn(rule).matches(pattern)]

    @classmethod
    def _match_complex_pattern_to_reaction_pattern(cls, pattern, test_pattern):
        for cp in test_pattern.complex_patterns:
            if match_complex_pattern(pattern, cp):
                return True
        return False

    def _cache_containing_monomers(self, cache, monomer_list):
        """
        Identifies rules containing a list of monomers

        Parameters
        ----------
        monomer_list: list of Monomers
            A list of monomers with which to search the model's rules

        Returns
        -------
        Model rules containing all of the monomers in the list

        """
        rule_names = set.intersection(*[cache[mon] for mon in
                                        monomer_list])
        return [r for r in self.model.rules if r.name in rule_names]


class ReactionPatternMatcher(object):
    """
    Match a pattern against a model's reactions list

    Methods are provided to match against reaction reactants, products or
    both. Searches can be Monomers, MonomerPatterns, ComplexPatterns or
    ReactionPatterns.

    Examples
    --------

    Create a PatternMatcher for the EARM 1.0 model

    >>> from pysb.examples.earm_1_0 import model
    >>> from pysb.bng import generate_equations
    >>> from pysb.pattern import ReactionPatternMatcher
    >>> generate_equations(model)
    >>> rpm = ReactionPatternMatcher(model)

    Assign some monomers to variables (only needed when importing a model
    instead of defining one interactively)

    >>> AMito, mCytoC, mSmac, cSmac = [model.monomers[m] for m in \
                                       ('AMito', 'mCytoC', 'mSmac', 'cSmac')]

    Search using a Monomer

    >>> rpm.match_products(mSmac) # doctest:+NORMALIZE_WHITESPACE
    [Rxn (reversible):
        Reactants: {'__s15': mSmac(b=None), '__s45': AMito(b=None)}
        Products: {'__s47': AMito(b=1) % mSmac(b=1)}
        Rate: __s15*__s45*kf21 - __s47*kr21
        Rules: [Rule('bind_mSmac_AMito', AMito(b=None) + mSmac(b=None) |
                AMito(b=1) % mSmac(b=1), kf21, kr21)]]

    Search using a MonomerPattern

    >>> rpm.match_reactants(AMito(b=ANY)) # doctest:+NORMALIZE_WHITESPACE
    [Rxn (one-way):
        Reactants: {'__s46': AMito(b=1) % mCytoC(b=1)}
        Products: {'__s45': AMito(b=None), '__s48': ACytoC(b=None)}
        Rate: __s46*kc20
        Rules: [Rule('produce_ACytoC_via_AMito', AMito(b=1) % mCytoC(b=1) >>
                AMito(b=None) + ACytoC(b=None), kc20)],
     Rxn (one-way):
        Reactants: {'__s47': AMito(b=1) % mSmac(b=1)}
        Products: {'__s45': AMito(b=None), '__s49': ASmac(b=None)}
        Rate: __s47*kc21
        Rules: [Rule('produce_ASmac_via_AMito', AMito(b=1) % mSmac(b=1) >>
                AMito(b=None) + ASmac(b=None), kc21)]]

    >>> rpm.match_products(cSmac(b=ANY)) # doctest:+NORMALIZE_WHITESPACE
    [Rxn (reversible):
        Reactants: {'__s7': XIAP(b=None), '__s51': cSmac(b=None)}
        Products: {'__s53': XIAP(b=1) % cSmac(b=1)}
        Rate: __s51*__s7*kf28 - __s53*kr28
        Rules: [Rule('inhibit_cSmac_by_XIAP', cSmac(b=None) + XIAP(b=None) |
                cSmac(b=1) % XIAP(b=1), kf28, kr28)]]

    Search using a ComplexPattern

    >>> rpm.match_reactants(AMito() % mSmac()) # doctest:+NORMALIZE_WHITESPACE
    [Rxn (one-way):
        Reactants: {'__s47': AMito(b=1) % mSmac(b=1)}
        Products: {'__s45': AMito(b=None), '__s49': ASmac(b=None)}
        Rate: __s47*kc21
        Rules: [Rule('produce_ASmac_via_AMito', AMito(b=1) % mSmac(b=1) >>
                AMito(b=None) + ASmac(b=None), kc21)]]

    >>> rpm.match_reactions(AMito(b=3) % mCytoC(b=3)) \
    # doctest:+NORMALIZE_WHITESPACE
    [Rxn (reversible):
        Reactants: {'__s14': mCytoC(b=None), '__s45': AMito(b=None)}
        Products: {'__s46': AMito(b=1) % mCytoC(b=1)}
        Rate: __s14*__s45*kf20 - __s46*kr20
        Rules: [Rule('bind_mCytoC_AMito', AMito(b=None) + mCytoC(b=None) |
                AMito(b=1) % mCytoC(b=1), kf20, kr20)],
     Rxn (one-way):
        Reactants: {'__s46': AMito(b=1) % mCytoC(b=1)}
        Products: {'__s45': AMito(b=None), '__s48': ACytoC(b=None)}
        Rate: __s46*kc20
        Rules: [Rule('produce_ACytoC_via_AMito', AMito(b=1) % mCytoC(b=1) >>
                AMito(b=None) + ACytoC(b=None), kc20)]]
    """
    def __init__(self, model, species_pattern_matcher=None):
        self.model = model

        # In this cache, our caches map species to reactions
        self._reactant_cache = collections.defaultdict(set)
        self._product_cache = collections.defaultdict(set)

        if not species_pattern_matcher:
            self.spm = SpeciesPatternMatcher(model)

        for r_id, rxn in enumerate(model.reactions_bidirectional):
            for cache, species_ids in (
                    (self._reactant_cache, rxn['reactants']),
                    (self._product_cache, rxn['products'])):
                for sp_id in species_ids:
                    sp = model.species[sp_id]
                    if sp.compartment:
                        raise NotImplementedError
                    cache[sp].add(r_id)

    def match_reactants(self, pattern):
        return self._match_reactions_against_cache(pattern, 'reactant')

    def match_products(self, pattern):
        return self._match_reactions_against_cache(pattern, 'product')

    def match_reactions(self, pattern):
        return self._match_reactions_against_cache(pattern, 'both')

    def _match_reactions_against_cache(self, pattern, reaction_side):
        species = self.spm.match(pattern)

        rxn_ids = set()
        if reaction_side in ['reactant', 'both']:
            rxn_ids.update(*[self._reactant_cache[sp] for sp in species])

        if reaction_side in ['product', 'both']:
            rxn_ids.update(*[self._product_cache[sp] for sp in species])
        rxn_ids = list(rxn_ids)
        rxn_ids.sort()

        return [_Reaction(rxn_dict=self.model.reactions_bidirectional[rxn_id],
                         model=self.model) for rxn_id in rxn_ids]


class _Reaction(object):
    __slots__ = ['_rxn_dict', 'reactants', 'model', 'products', 'species']
    """
    Store reactions in object form for pretty-printing
    """
    def __init__(self, rxn_dict=None, model=None, species=None):
        self._rxn_dict = rxn_dict

        if model is None:
            raise ValueError('Must specify model or species list')

        self.model = model

        if species:
            self.species = species
        else:
            self.species = model.species

        self.reactants = collections.defaultdict(int)
        self.products = collections.defaultdict(int)

        for r_id in rxn_dict['reactants']:
            self.reactants[r_id] += 1

        for p_id in rxn_dict['products']:
            self.products[p_id] += 1

    @property
    def reversible(self):
        return self._rxn_dict.get('reversible', None)

    @property
    def reverse(self):
        return self._rxn_dict.get('reverse', None)

    @property
    def rules(self):
        return [self.model.rules[r] for r in self._rxn_dict['rule']]

    def add_rule(self, rule):
        if rule.name not in self._rxn_dict['rule']:
            self._rxn_dict['rule'].append(rule.name)

    @property
    def rate(self):
        return self._rxn_dict['rate']

    def __repr__(self):
        return 'Rxn (%s): \n    Reactants: %s\n    Products: %s\n    ' \
               'Rate: %s\n    Rules: %s' % (
                    'reversible' if self.reversible else
                    ('one-way [reverse]' if self.reverse else
                        'one-way'),
                    self._repr_species_dict(self.reactants),
                    self._repr_species_dict(self.products),
                    self.rate,
                    self.rules
               )

    def __cmp__(self, other):
        try:
            return self._rxn_dict == other._rxn_dict
        except AttributeError:
            return False

    def _repr_species_dict(self, species_dict):
        return '{%s}' % ', '.join(["'__s%d': %s" % (k, self.species[k])
                                   for k in sorted(species_dict.keys())])



from pysb.core import MonomerPattern, ComplexPattern, RuleExpression, \
    ReactionPattern, ANY, WILD
from pysb.builder import Builder
from pysb.bng import BngFileInterface
import xml.etree.ElementTree
import re
from sympy.parsing.sympy_parser import parse_expr
import warnings


def _ns(tag_string):
    """
    Shorthand to apply XML namespace
    """
    return tag_string.format('{http://www.sbml.org/sbml/level3}')


class BnglImportError(Exception):
    pass


class BnglBuilder(Builder):
    """
    Assemble a Model from a .bngl file.

    See :py:func:`model_from_bngl` for further details.
    """
    def __init__(self, filename, force=False, cleanup=True):
        super(BnglBuilder, self).__init__()
        with BngFileInterface(model=None, cleanup=cleanup) as con:
            con.action('readFile', file=filename, skip_actions=1)
            con.action('writeXML', evaluate_expressions=0)
            con.execute()
            self._force = force
            self._x = xml.etree.ElementTree.parse('%s.xml' %
                                                  con.base_filename)\
                                           .getroot().find(_ns('{0}model'))
            self._model_env = {}
        self._parse_bng_xml()

    def _warn_or_except(self, msg):
        """
        Raises a warning or Exception, depending of the value of self._force
        """
        if self._force:
            warnings.warn(msg)
        else:
            raise BnglImportError(msg)

    def _eval_in_model_env(self, expression):
        """
        Evaluates an expression string in the model environment
        """
        components = dict((c.name, c) for c in self.model.all_components())
        self._model_env.update(components)

        # Quick security check on the expression
        if re.match(r'^[\w\s()/+\-._*]*$', expression):
            return eval(expression, {}, self._model_env)
        else:
            self._warn_or_except('Security check on expression "%s" failed' %
                                 expression)
            return None

    def _parse_species(self, species_xml):
        # Species may be empty for synthesis/degradation reaction patterns
        if species_xml is None:
            return []
        # Give the bonds unique numeric IDs
        bond_ids = {}
        next_bond_id = 1
        for bond in species_xml.iterfind(_ns('{0}ListOfBonds/{0}Bond')):
            for bond_attr in bond.attrib.keys():
                if bond_attr.startswith('site'):
                    bond_ids.setdefault(bond.attrib[bond_attr], [])\
                                        .append(next_bond_id)
            next_bond_id += 1

        # Create a list of monomer patterns
        mon_pats = []
        for mon in species_xml.iterfind(_ns('{0}ListOfMolecules/{0}Molecule')):
            mon_name = mon.get('name')
            mon_obj = self.model.monomers[mon_name]
            mon_states = {}
            for comp in mon.iterfind(_ns('{0}ListOfComponents/{0}Component')):
                state_nm = comp.get('name')
                bonds = comp.get('numberOfBonds')
                if bonds == "0":
                    mon_states[state_nm] = None
                elif bonds == "?":
                    mon_states[state_nm] = WILD
                elif bonds == "+":
                    mon_states[state_nm] = ANY
                else:
                    bond_list = bond_ids[comp.get('id')]
                    assert int(bonds) == len(bond_list)
                    if len(bond_list) == 1:
                        mon_states[state_nm] = bond_list[0]
                    else:
                        mon_states[state_nm] = bond_list
                state = comp.get('state')
                if state:
                    if mon_states[state_nm]:
                        # Site has a bond and a state
                        mon_states[state_nm] = (state, mon_states[state_nm])
                    else:
                        # Site only has a state, no bond
                        mon_states[state_nm] = state
            mon_cpt = self.model.compartments.get(mon.get('compartment'))
            mon_pats.append(MonomerPattern(mon_obj, mon_states, mon_cpt))
        return mon_pats

    def _parse_monomers(self):
        for m in self._x.iterfind(_ns('{0}ListOfMoleculeTypes/'
                                      '{0}MoleculeType')):
            mon_name = m.get('id')
            sites = []
            states = {}
            for ctype in m.iterfind(_ns('{0}ListOfComponentTypes/'
                                        '{0}ComponentType')):
                c_name = ctype.get('id')
                sites.append(c_name)
                states_list = ctype.find(_ns('{}ListOfAllowedStates'))
                if states_list is not None:
                    states[c_name] = [s.get('id') for s in
                                      states_list.iterfind(_ns(
                                                           '{}AllowedState'))]
            try:
                self.monomer(mon_name, sites, states)
            except Exception as e:
                if str(e).startswith('Duplicate sites specified'):
                    self._warn_or_except('Molecule %s has multiple '
                                         'sites with the same name. '
                                         'This is not supported in PySB.' %
                                         mon_name)
                else:
                    raise BnglImportError(str(e))

    def _parse_parameters(self):
        for p in self._x.iterfind(_ns('{0}ListOfParameters/{0}Parameter')):
            p_name = p.get('id')
            if p.get('type') == 'Constant':
                p_value = p.get('value').replace('10^', '1e')
                self.parameter(name=p_name, value=p_value)
            elif p.get('type') == 'ConstantExpression':
                self.expression(name=p_name,
                                expr=self._eval_in_model_env(p.get('value')))
            else:
                self._warn_or_except('Parameter %s has unknown type: %s' %
                                     (p_name, p.get('type')))

    def _parse_observables(self):
        for o in self._x.iterfind(_ns('{0}ListOfObservables/{0}Observable')):
            o_name = o.get('name')
            cplx_pats = []
            for mp in o.iterfind(_ns('{0}ListOfPatterns/{0}Pattern')):
                match_once = mp.get('matchOnce')
                match_once = True if match_once == "1" else False
                cplx_pats.append(ComplexPattern(self._parse_species(mp),
                                                compartment=None,
                                                match_once=match_once))
            self.observable(o_name,
                            ReactionPattern(cplx_pats),
                            match=o.get('type').lower())

    def _parse_initials(self):
        for i in self._x.iterfind(_ns('{0}ListOfSpecies/{0}Species')):
            if i.get('Fixed') is not None and i.get('Fixed') == "1":
                self._warn_or_except('Species %s is fixed, but will be '
                                     'treated as an ordinary species in '
                                     'PySB.' % i.get('name'))

            value_param = i.get('concentration')
            try:
                value = float(value_param)
                # Need to create a new parameter for the initial conc. literal
                name = re.sub('[^\w]+', '_', i.get('name').replace(
                    ')', '').replace('(', '')) + '_0'
                try:
                    value_param = self.parameter(name, value)
                except ValueError as ve:
                    raise BnglImportError(str(ve))
            except ValueError:
                # Retrieve existing parameter or (constant) expression
                try:
                    value_param = self.model.parameters[i.get('concentration')]
                except KeyError:
                    value_param = self.model.expressions[i.get(
                        'concentration')]
            mon_pats = self._parse_species(i)
            species_cpt = self.model.compartments.get(i.get('compartment'))
            self.initial(ComplexPattern(mon_pats, species_cpt), value_param)

    def _parse_compartments(self):
        for c in self._x.iterfind(_ns('{0}ListOfCompartments/{0}compartment')):
            cpt_size = None
            if c.get('size'):
                cpt_size = self.parameter('%s_size' % c.get('id'),
                                          c.get('size'))
            cpt_parent = None
            if c.get('outside'):
                cpt_parent = self.model.compartments[c.get('outside')]
            self.compartment(name=c.get('id'),
                             parent=cpt_parent,
                             dimension=int(c.get('spatialDimensions')),
                             size=cpt_size)

    def _parse_rate_law(self, rl):
        if rl.get('type') == 'Ele':
            rate_law_name = rl.find(_ns('{0}ListOfRateConstants/'
                                        '{0}RateConstant')).get('value')
            try:
                return self.model.parameters[rate_law_name]
            except KeyError:
                return self.model.expressions[rate_law_name]
        elif rl.get('type') == 'Function':
            return self.model.expressions[rl.get('name')]
        else:
            self._warn_or_except('Rate law %s has unknown type %s' %
                                 (rl.get('id'), rl.get('type')))
            return None

    def _parse_rules(self):
        # Store reversible rates for post-processing (we don't know if we'll
        # encounter fwd or rev rule first)
        rev_rates = {}
        for r in self._x.iterfind(_ns('{0}ListOfReactionRules/'
                                      '{0}ReactionRule')):
            r_name = r.get('name')

            r_rate_xml = r.find(_ns('{}RateLaw'))
            if r_rate_xml is None:
                raise BnglImportError('Rate law missing for rule %s' % r_name)
            r_rate = self._parse_rate_law(r_rate_xml)

            # Store reverse rates for post-processing
            if r_name.startswith('_reverse_'):
                r_name = r_name[9:]
                rev_rates[r_name] = r_rate
                continue

            # Compile reactant and product patterns
            reactant_pats = []
            for rp in r.iterfind(_ns('{0}ListOfReactantPatterns/'
                                     '{0}ReactantPattern')):
                cpt = self.model.compartments.get(rp.get('compartment'))
                reactant_pats.append(ComplexPattern(self._parse_species(rp),
                                                    cpt))
            product_pats = []
            for pp in r.iterfind(_ns('{0}ListOfProductPatterns/'
                                     '{0}ProductPattern')):
                cpt = self.model.compartments.get(pp.get('compartment'))
                product_pats.append(ComplexPattern(self._parse_species(pp),
                                                   cpt))
            rule_exp = RuleExpression(ReactionPattern(reactant_pats),
                                      ReactionPattern(product_pats),
                                      is_reversible=False)

            # Process any DeleteMolecules declaration
            delete_molecules = False
            for del_operations in r.iterfind(_ns('{0}ListOfOperations/'
                                                 '{0}Delete')):
                if del_operations.get('DeleteMolecules') == "1":
                    delete_molecules = True
                    break

            # Give warning/error if ListOfExcludeReactants or
            # ListOfExcludeProducts is present
            if r.find(_ns('{}ListOfExcludeReactants')) is not None or \
               r.find(_ns('{}ListOfExcludeProducts')) is not None:
                self._warn_or_except('ListOfExcludeReactants and/or '
                                     'ListOfExcludeProducts declarations will '
                                     'be ignored. This may lead to long '
                                     'network generation times.')

            # Give warning/error if ListOfIncludeReactants or
            # ListOfIncludeProducts is present
            if r.find(_ns('{}ListOfIncludeReactants')) is not None or \
               r.find(_ns('{}ListOfIncludeProducts')) is not None:
                self._warn_or_except('ListOfIncludeReactants and/or '
                                     'ListOfIncludeProducts declarations will '
                                     'be ignored. This may lead to long '
                                     'network generation times.')

            self.rule(r_name, rule_exp, r_rate,
                      delete_molecules=delete_molecules)

        # Set the reverse rates
        for r_name, rev_rate in rev_rates.items():
            rule = self.model.rules[r_name]
            rule.rule_expression.is_reversible = True
            rule.is_reversible = True
            rule.rate_reverse = rev_rate

    def _parse_expressions(self):
        expr_namespace = {p.name: p.value for p in self.model.parameters}
        expr_namespace.update({o.name: o for o in
                               self.model.observables})

        for e in self._x.iterfind(_ns('{0}ListOfFunctions/{0}Function')):
            if e.find(_ns('{0}ListOfArguments/{0}Argument')) is not None:
                self._warn_or_except('Function %s is local, which is not '
                                     'supported in PySB' % e.get('id'))
            expr_name = e.get('id')
            expr_text = e.find(_ns('{0}Expression')).text.replace('^', '**')
            expr_val = 0
            try:
                expr_val = parse_expr(expr_text, local_dict=expr_namespace)
            except Exception as ex:
                self._warn_or_except('Could not parse expression %s: '
                                     '%s\n\nError: %s' % (expr_name,
                                                          expr_text,
                                                          ex.message))
            expr_namespace[expr_name] = expr_val
            self.expression(expr_name, expr_val)

    def _parse_bng_xml(self):
        self.model.name = self._x.get(_ns('id'))
        self._parse_monomers()
        self._parse_parameters()
        self._parse_compartments()
        self._parse_initials()
        self._parse_observables()
        self._parse_expressions()
        self._parse_rules()


def model_from_bngl(filename, force=False, cleanup=True):
    """
    Convert a BioNetGen .bngl model file into a PySB Model.

    Notes
    -----

    The following features are not supported in PySB and will cause an error
    if present in a .bngl file:

    * Fixed species (with a ``$`` prefix, like ``$Null``)
    * BNG excluded or included reaction patterns (deprecated in BNG)
    * BNG local functions
    * Molecules with identically named sites, such as ``M(l,l)``
    * BNG's custom rate law functions, such as ``MM`` and ``Sat``
      (deprecated in BNG)

    Parameters
    ----------
    filename : string
        A BioNetGen .bngl file
    force : bool, optional
        The default, False, will raise an Exception if there are any errors
        importing the model to PySB, e.g. due to unsupported features.
        Setting to True will attempt to ignore any import errors, which may
        lead to a model that only poorly represents the original. Use at own
        risk!
    cleanup : bool
        Delete temporary directory on completion if True. Set to False for
        debugging purposes.
    """
    bb = BnglBuilder(filename, force=force, cleanup=cleanup)
    return bb.model



from nose.tools import *
from pysb.core import Model, SelfExporter
import pickle

def with_model(func):
    """Decorate a test to set up and tear down a Model."""
    def inner(*args, **kwargs):
        model = Model(func.__name__, _export=False)
        # manually set up SelfExporter, targeting func's globals
        selfexporter_state = SelfExporter.do_export
        SelfExporter.do_export = True
        SelfExporter.default_model = model
        SelfExporter.target_module = func.__module__
        SelfExporter.target_globals = func.__globals__
        SelfExporter.target_globals['model'] = model
        try:
            # call the actual test function
            func(*args, **kwargs)
        finally:
            # clean up the globals
            SelfExporter.cleanup()
            SelfExporter.do_export = selfexporter_state
    return make_decorator(func)(inner)

def serialize_component_list(model, filename):
    """Serialize (pickle) the components of the given model to a file. This can
    later be used to compare the state of the model against a previously
    validated state using :py:func:`check_model_against_component_list`.
    """

    f = open(filename, 'w')
    pickle.dump(list(model.all_components().values()), f)
    f.close()

def check_model_against_component_list(model, component_list):
    """Check the components of the given model against the provided list
    of components, asserting that they are equal. Useful for testing a
    model against a previously validated (and serialized) state.

    Currently checks equality by performing a string comparison of the
    repr() of each component, however, this may be revised to use alternative
    measures of equality in the future.
    
    To serialize the list of components to create a record of a
    validated state, see :py:func:`serialize_component_list`.
    """
    assert len(model.all_components()) == len(component_list), \
           "Model %s does not have the same " \
           "number of components as the previously validated version. " \
           "The validated model has %d components, current model has " \
           "%d components." % \
           (model.name, len(model.all_components()), len(component_list))

    model_components = list(model.all_components().values())
    for i, comp in enumerate(component_list):
        model_comp_str = repr(model_components[i])
        comp_str = repr(comp) 
        assert comp_str == model_comp_str, \
               "Model %s does not match reference version: " \
               "Mismatch at component %d: %s in the reference model not " \
               "equal to %s in the current model." \
                % (model.name, i, comp_str, model_comp_str)

    assert True



import sys
import os
import errno
import warnings
import inspect
import re
import collections
import weakref
import copy
import itertools
import sympy
import scipy.sparse
import networkx as nx

try:
    reload
except NameError:
    from imp import reload
try:
    basestring
except NameError:
    # Under Python 3, do not pretend that bytes are a valid string
    basestring = str
    long = int

def Initial(*args):
    """Declare an initial condition (see Model.initial)."""
    return SelfExporter.default_model.initial(*args)

def MatchOnce(pattern):
    """Make a ComplexPattern match-once."""
    cp = as_complex_pattern(pattern).copy()
    cp.match_once = True
    return cp


# A module may define a global with this name (_pysb_doctest_...) to request
# that SelfExporter not issue any ModelExistsWarnings from doctests defined
# therein. (This is the best method we could come up with to manage this
# behavior, as doctest doesn't offer per-doctest setup/teardown.)
_SUPPRESS_MEW = '_pysb_doctest_suppress_modelexistswarning'

class SelfExporter(object):

    """
    Make model components appear in the calling module's namespace.

    This class is for pysb internal use only. Do not construct any instances.

    """
    
    do_export = True
    default_model = None
    target_globals = None   # the globals dict to which we'll export our symbols
    target_module = None    # the module to which we've exported

    @staticmethod
    def export(obj):
        """Export an object by name and add it to the default model."""

        if not SelfExporter.do_export:
            return
        if not isinstance(obj, (Model, Component)):
            raise Exception("%s is not a type that is understood by SelfExporter" % str(type(obj)))

        # determine the module from which we were called (we need to do this here so we can
        # calculate stacklevel for use in the warning at the bottom of this method)
        cur_module = inspect.getmodule(inspect.currentframe())
        caller_frame = inspect.currentframe()
        # walk up through the stack until we hit a different module
        stacklevel = 1
        while inspect.getmodule(caller_frame) == cur_module:
            stacklevel += 1
            caller_frame = caller_frame.f_back

        # use obj's name as the symbol to export it to (unless modified below)
        export_name = obj.name

        if isinstance(obj, Model):
            new_target_module = inspect.getmodule(caller_frame)
            if SelfExporter.default_model is not None \
                    and new_target_module is SelfExporter.target_module:
                # Warn, unless running a doctest whose containing module set the
                # magic global which tells us to suppress it.
                if not (
                    caller_frame.f_code.co_filename.startswith('<doctest ') and
                    caller_frame.f_globals.get(_SUPPRESS_MEW)):
                    warnings.warn("Redefining model! (You can probably ignore "
                                  "this if you are running code interactively)",
                                  ModelExistsWarning, stacklevel)
                SelfExporter.cleanup()
            SelfExporter.target_module = new_target_module
            SelfExporter.target_globals = caller_frame.f_globals
            SelfExporter.default_model = obj
            # if not set, assign model's name from the module it lives in. very sneaky and fragile.
            if obj.name is None:
                if SelfExporter.target_module == sys.modules['__main__']:
                    # user ran model .py directly
                    model_path = inspect.getfile(sys.modules['__main__'])
                    model_filename = os.path.basename(model_path)
                    module_name = re.sub(r'\.py$', '', model_filename)
                elif SelfExporter.target_module is not None:
                    # model is imported by some other script (typical case)
                    module_name = SelfExporter.target_module.__name__
                else:
                    # user is defining a model interactively (not really supported, but we'll try)
                    module_name = '_interactive_'
                obj.name = module_name   # internal name for identification
                export_name = 'model'    # symbol name for export
        elif isinstance(obj, Component):
            if SelfExporter.default_model == None:
                raise Exception("A Model must be declared before declaring any model components")
            SelfExporter.default_model.add_component(obj)

        # load obj into target namespace under obj.name
        if export_name in SelfExporter.target_globals:
            warnings.warn("'%s' already defined" % (export_name), SymbolExistsWarning, stacklevel)
        SelfExporter.target_globals[export_name] = obj

    @staticmethod
    def cleanup():
        """Delete previously exported symbols."""
        if SelfExporter.default_model is None:
            return
        for name in [c.name for c in SelfExporter.default_model.all_components()] + ['model']:
            if name in SelfExporter.target_globals:
                del SelfExporter.target_globals[name]
        SelfExporter.default_model = None
        SelfExporter.target_globals = None
        SelfExporter.target_module = None

    @staticmethod
    def rename(obj, new_name):
        """Rename a previously exported symbol"""
        if new_name in SelfExporter.target_globals:
            msg = "'%s' already defined" % new_name
            warnings.warn(msg, SymbolExistsWarning, 2)
        if obj.name in SelfExporter.target_globals:
            obj = SelfExporter.target_globals[obj.name]
            SelfExporter.target_globals[new_name] = obj
            del SelfExporter.target_globals[obj.name]
        else:
            raise ValueError("Could not find object in global namespace by its"
                             "name '%s'" % obj.name)


class Component(object):

    """
    The base class for all the named things contained within a model.

    Parameters
    ----------
    name : string
        Name of the component. Must be unique within the containing model.

    Attributes
    ----------
    name : string
        Name of the component.
    model : weakref(Model)
        Containing model.

    """

    def __init__(self, name, _export=True):
        if not re.match(r'[_a-z][_a-z0-9]*\Z', name, re.IGNORECASE):
            raise InvalidComponentNameError(name)
        self.name = name
        self.model = None  # to be set in Model.add_component
        self._export = _export
        if self._export:
            self._do_export()

    def __getstate__(self):
        # clear the weakref to parent model (restored in Model.__setstate__)
        state = self.__dict__.copy()
        del state['model']
        # Force _export to False; we don't want the unpickling process to
        # trigger SelfExporter.export!
        state['_export'] = False
        return state

    def _do_export(self):
        try:
            SelfExporter.export(self)
        except ComponentDuplicateNameError as e:
            # re-raise to hide the stack trace below this point -- it's irrelevant to the user
            # and makes the error harder to understand
            raise e

    def rename(self, new_name):
        """Change component's name.

        This is typically only needed when deriving one model from another and
        it would be desirable to change a component's name in the derived
        model."""
        if self.model:
            self.model()._rename_component(self, new_name)
        if self._export:
            SelfExporter.rename(self, new_name)
        self.name = new_name


class Monomer(Component):

    """
    Model component representing a protein or other molecule.

    Parameters
    ----------
    sites : list of strings, optional
        Names of the sites.
    site_states : dict of string => string, optional
        Allowable states for sites. Keys are sites and values are lists of
        states. Sites which only take part in bond formation and never take on a
        state may be omitted.

    Attributes
    ----------
    Identical to Parameters (see above).

    Notes
    -----

    A Monomer instance may be \"called\" like a function to produce a
    MonomerPattern, as syntactic sugar to approximate rule-based modeling
    language syntax. It is typically called with keyword arguments where the arg
    names are sites and values are site conditions such as bond numbers or
    states (see the Notes section of the :py:class:`MonomerPattern`
    documentation for details). To help in situations where kwargs are unwieldy
    (for example if a site name is computed dynamically or stored in a variable)
    a dict following the same layout as the kwargs may be passed as the first
    and only positional argument instead.

    """

    def __init__(self, name, sites=None, site_states=None, _export=True):
        Component.__init__(self, name, _export)

        # Create default empty containers.
        if sites is None:
            sites = []
        if site_states is None:
            site_states = {}

        # ensure sites is some kind of list (presumably of strings) but not a
        # string itself
        if not isinstance(sites, collections.Iterable) or \
               isinstance(sites, basestring):
            raise ValueError("sites must be a list of strings")

        # ensure no duplicate sites
        sites_seen = {}
        for site in sites:
            sites_seen.setdefault(site, 0)
            sites_seen[site] += 1
        sites_dup = [site for site, count in sites_seen.items() if count > 1]
        if sites_dup:
            raise Exception("Duplicate sites specified: " + str(sites_dup))

        # ensure site_states keys are all known sites
        unknown_sites = [site for site in site_states if not site in sites_seen]
        if unknown_sites:
            raise Exception("Unknown sites in site_states: " +
                            str(unknown_sites))
        # ensure site_states values are all strings
        invalid_sites = [site for (site, states) in site_states.items()
                              if not all([isinstance(s, basestring)
                                          for s in states])]
        if invalid_sites:
            raise Exception("Non-string state values in site_states for "
                            "sites: " + str(invalid_sites))

        self.sites = list(sites)
        self.site_states = site_states

    def __call__(self, conditions=None, **kwargs):
        """
        Return a MonomerPattern object based on this Monomer.

        See the Notes section of this class's documentation for details.

        Parameters
        ----------
        conditions : dict, optional
            See MonomerPattern.site_conditions.
        **kwargs : dict
            See MonomerPattern.site_conditions.

        """
        return MonomerPattern(self, extract_site_conditions(conditions,
                                                            **kwargs), None)

    def __repr__(self):
        value = '%s(%s' % (self.__class__.__name__, repr(self.name))
        if self.sites:
            value += ', %s' % repr(self.sites)
        if self.site_states:
            value += ', %s' % repr(self.site_states)
        value += ')'
        return value


def _check_state(monomer, site, state):
    """ Check a monomer site allows the specified state """
    if state not in monomer.site_states[site]:
        args = state, monomer.name, site, monomer.site_states[site]
        template = "Invalid state choice '{}' in Monomer {}, site {}. Valid " \
                   "state choices: {}"
        raise ValueError(template.format(*args))


class MonomerPattern(object):

    """
    A pattern which matches instances of a given monomer.

    Parameters
    ----------
    monomer : Monomer
        The monomer to match.
    site_conditions : dict
        The desired state of the monomer's sites. Keys are site names and values
        are described below in Notes.
    compartment : Compartment or None
        The desired compartment where the monomer should exist. None means
        \"don't-care\".

    Attributes
    ----------
    Identical to Parameters (see above).

    Notes
    -----
    The acceptable values in the `site_conditions` dict are as follows:

    * ``None`` : no bond
    * *str* : state
    * *int* : a bond (to a site with the same number in a ComplexPattern)
    * *list of int* : multi-bond (not valid in Kappa)
    * ``ANY`` : \"any\" bond (bound to something, but don't care what)
    * ``WILD`` : \"wildcard\" bond (bound or not bound)
    * *tuple of (str, int)* : state with specified bond
    * *tuple of (str, WILD)* : state with wildcard bond
    * *tuple of (str, ANY)* : state with any bond

    If a site is not listed in site_conditions then the pattern will match any
    state for that site, i.e. \"don't write, don't care\".

    """
    def __init__(self, monomer, site_conditions, compartment):
        # ensure all keys in site_conditions are sites in monomer
        unknown_sites = [site for site in site_conditions
                              if site not in monomer.sites]
        if unknown_sites:
            raise Exception("MonomerPattern with unknown sites in " +
                            str(monomer) + ": " + str(unknown_sites))

        # ensure each value is one of: None, integer, list of integers, string,
        # (string,integer), (string,WILD), ANY, WILD
        invalid_sites = []
        for (site, state) in site_conditions.items():
            # pass through to next iteration if state type is ok
            if state == None:
                continue
            elif isinstance(state, int):
                continue
            elif isinstance(state, list) and \
                 all(isinstance(s, int) for s in state):
                continue
            elif isinstance(state, basestring):
                _check_state(monomer, site, state)
                continue
            elif isinstance(state, tuple) and \
                 isinstance(state[0], basestring) and \
                 (isinstance(state[1], int) or state[1] is WILD or \
                  state[1] is ANY):
                _check_state(monomer, site, state[0])
                continue
            elif state is ANY:
                continue
            elif state is WILD:
                continue
            invalid_sites.append(site)
        if invalid_sites:
            raise Exception("Invalid state value for sites: " +
                            '; '.join(['%s=%s' % (s,str(site_conditions[s]))
                                       for s in invalid_sites]))

        # ensure compartment is a Compartment
        if compartment and not isinstance(compartment, Compartment):
            raise Exception("compartment is not a Compartment object")

        self.monomer = monomer
        self.site_conditions = site_conditions
        self.compartment = compartment
        self._graph = None

    def is_concrete(self):
        """
        Return a bool indicating whether the pattern is 'concrete'.

        'Concrete' means the pattern satisfies ALL of the following:

        1. All sites have specified conditions
        2. If the model uses compartments, the compartment is specified.

        """
        # 1.
        sites_ok = self.is_site_concrete()
        # 2.
        compartment_ok = not self.monomer.model().compartments or self.compartment
        return compartment_ok and sites_ok

    def is_site_concrete(self):
        """
        Return a bool indicating whether the pattern is 'site-concrete'.

        'Site-concrete' means all sites have specified conditions."""
        if len(self.site_conditions) != len(self.monomer.sites):
            return False
        for site_name, site_val in self.site_conditions.items():
            if isinstance(site_val, basestring):
                site_state = site_val
                site_bond = None
            elif isinstance(site_val, collections.Iterable):
                site_state, site_bond = site_val
            elif isinstance(site_val, int):
                site_bond = site_val
                site_state = None
            else:
                site_bond = site_val
                site_state = None

            if site_bond is ANY or site_bond is WILD:
                return False
            if site_state is None and site_name in \
                    self.monomer.site_states.keys():
                return False

        return True

    def _as_graph(self):
        """
        Convert MonomerPattern to networkx graph, caching the result

        See :func:`ComplexPattern._as_graph` for implementation details
        """
        if self._graph is None:
            self._graph = as_complex_pattern(self)._as_graph()

        return self._graph

    def __call__(self, conditions=None, **kwargs):
        """Build a new MonomerPattern with updated site conditions. Can be used
        to obtain a shallow copy by passing an empty argument list."""
        # The new object will have references to the original monomer and
        # compartment, and a shallow copy of site_conditions which has been
        # updated according to our args (as in Monomer.__call__).
        site_conditions = self.site_conditions.copy()
        site_conditions.update(extract_site_conditions(conditions, **kwargs))
        return MonomerPattern(self.monomer, site_conditions, self.compartment)

    def __add__(self, other):
        if isinstance(other, MonomerPattern):
            return ReactionPattern([ComplexPattern([self], None), ComplexPattern([other], None)])
        if isinstance(other, ComplexPattern):
            return ReactionPattern([ComplexPattern([self], None), other])
        elif other is None:
            rp = as_reaction_pattern(self)
            rp.complex_patterns.append(None)
            return rp
        else:
            return NotImplemented

    def __radd__(self, other):
        if other is None:
            rp = as_reaction_pattern(self)
            rp.complex_patterns = [None] + rp.complex_patterns
            return rp
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, MonomerPattern):
            return ComplexPattern([self, other], None)
        else:
            return NotImplemented

    def __rshift__(self, other):
        return build_rule_expression(self, other, False)

    def __rrshift__(self, other):
        return build_rule_expression(other, self, False)

    def __or__(self, other):
        return build_rule_expression(self, other, True)

    def __ne__(self, other):
        warnings.warn("'<>' for reversible rules will be removed in a future "
                      "version of PySB. Use '|' instead.",
                      PendingDeprecationWarning,
                      stacklevel=2)
        return self.__or__(other)

    def __pow__(self, other):
        if isinstance(other, Compartment):
            if self.compartment is not None:
                raise CompartmentAlreadySpecifiedError()
            mp_new = self()
            mp_new.compartment = other
            return mp_new
        else:
            return NotImplemented

    def __repr__(self):
        value = '%s(' % self.monomer.name
        value += ', '.join([
                k + '=' + repr(self.site_conditions[k])
                for k in self.monomer.sites
                if k in self.site_conditions
                ])
        value += ')'
        if self.compartment is not None:
            value += ' ** ' + self.compartment.name
        return value



class ComplexPattern(object):

    """
    A bound set of MonomerPatterns, i.e. a pattern to match a complex.

    In BNG terms, a list of patterns combined with the '.' operator.

    Parameters
    ----------
    monomer_patterns : list of MonomerPatterns
        MonomerPatterns that make up the complex.
    compartment : Compartment or None
        Location restriction. None means don't care.
    match_once : bool, optional
        If True, the pattern will only count once against a species in which the
        pattern can match the monomer graph in multiple distinct ways. If False
        (default), the pattern will count as many times as it matches the
        monomer graph, leading to a faster effective reaction rate.

    Attributes
    ----------
    Identical to Parameters (see above).

    """

    def __init__(self, monomer_patterns, compartment, match_once=False):
        # ensure compartment is a Compartment
        if compartment and not isinstance(compartment, Compartment):
            raise Exception("compartment is not a Compartment object")

        self.monomer_patterns = monomer_patterns
        self.compartment = compartment
        self.match_once = match_once
        self._graph = None

    def is_concrete(self):
        """
        Return a bool indicating whether the pattern is 'concrete'.

        'Concrete' means the pattern satisfies ANY of the following:
        1. All monomer patterns are concrete
        2. The compartment is specified AND all monomer patterns are site-concrete
        """
        # 1.
        mp_concrete_ok = all(mp.is_concrete() for mp in self.monomer_patterns)
        # 2.
        compartment_ok = self.compartment is not None and \
            all(mp.is_site_concrete() for mp in self.monomer_patterns)
        return mp_concrete_ok or compartment_ok

    def _as_graph(self):
        """
        Return the ComplexPattern represented as a networkx graph

        ComplexPatterns can be represented as a graph. This is mainly useful
        for comparing if ComplexPatterns are equivalent (see
        :func:`ComplexPattern.is_equivalent_to`).

        It turns out this is non-trivial because 1) bond numbering is
        arbitrary and 2) ComplexPatterns can contain MonomerPatterns which
        are identical. The latter problem makes it impossible to merely
        order the MonomerPatterns using a canonical ordering for comparison,
        while ensuring correctness in all cases [Blinov2006]_.

        We solve the problem using broadly the same approach as BioNetGen -
        encode each complex pattern as a graph and check if they are
        isomorphic to each other [Faeder2009]_. However, our approach
        differs in that we do not need to use a hierarchical graph like
        BioNetGen's hnauty algorithm. We use networkx, in which graph nodes are
        Python objects rather than strings; thus, we ensure that
        monomers/sites/states with the same name are not evaluated to be
        equal, because they have different object type.

        **Implementation details**
        Each monomer, site, state and compartment is represented as a node.
        Edges represent bonds (when between sites), or a relationship
        (monomers have sites, sites have states, MonomerPatterns and
        ComplexPatterns can have Compartments). A special "no bond" node is
        used to denote that the connected site is unbound; this is necessary
        because pattern matching is performed by checking for an isomorphic
        subgraph, and we need to distinguish between explicitly unbound and
        unspecified bond (equivalent to the `ANY` keyword).

        Internally, networkx references nodes using an integer. We use a
        private autoincrementing integer generator function `autoinc` to track
        nodes, but this is not used when checking graph isomorphism (instead,
        node to node object equality is checked).

        The `WILD` keyword should match any bond except the special "no
        bond" node - as special private `WildTester` function is used for
        this purpose.

        Compartment nodes are tracked and kept unique by the private
        `add_or_get_compartment_node` function, which uses a dictionary to
        track Compartment->node_id mapping.

        .. [Blinov2006] https://link.springer.com/chapter/10.1007%2F11905455_5
        .. [Faeder2009] https://www.csb.pitt.edu/Faculty/Faeder/Publications/Reprints/Faeder_2009.pdf
        """
        if self._graph is not None:
            return self._graph

        NO_BOND = 'NoBond'

        def autoinc():
            i = 0
            while True:
                yield i
                i += 1
        node_count = autoinc()

        class AnyBondTester(object):
            def __eq__(self, other):
                return not isinstance(other, Component) and other != NO_BOND

        any_bond_tester = AnyBondTester()

        bond_edges = collections.defaultdict(list)
        g = nx.Graph()
        _cpt_nodes = {}

        def add_or_get_compartment_node(cpt):
            try:
                return _cpt_nodes[cpt]
            except KeyError:
                cpt_node_id = next(node_count)
                _cpt_nodes[cpt] = cpt_node_id
                g.add_node(cpt_node_id, id=cpt)
                return cpt_node_id

        species_cpt_node_id = None
        if self.compartment:
            species_cpt_node_id = add_or_get_compartment_node(self.compartment)

        for mp in self.monomer_patterns:
            mon_node_id = next(node_count)
            g.add_node(mon_node_id, id=mp.monomer)
            if mp.compartment or self.compartment:
                cpt_node_id = add_or_get_compartment_node(mp.compartment or
                                                          self.compartment)
                g.add_edge(mon_node_id, cpt_node_id)

            for site, state_or_bond in mp.site_conditions.items():
                mon_site_id = next(node_count)
                g.add_node(mon_site_id, id=site)
                g.add_edge(mon_node_id, mon_site_id)
                state = None
                bond_num = None
                if state_or_bond is WILD:
                    continue
                elif isinstance(state_or_bond, basestring):
                    state = state_or_bond
                elif isinstance(state_or_bond, collections.Iterable) and len(
                        state_or_bond) == 2:
                    state = state_or_bond[0]
                    bond_num = state_or_bond[1]
                elif isinstance(state_or_bond, int):
                    bond_num = state_or_bond

                if state_or_bond is ANY or bond_num is ANY:
                    bond_num = any_bond_tester
                    any_bond_tester_id = next(node_count)
                    g.add_node(any_bond_tester_id, id=any_bond_tester)
                    g.add_edge(mon_site_id, any_bond_tester_id)

                if state is not None:
                    mon_site_state_id = next(node_count)
                    g.add_node(mon_site_state_id, id=state)
                    g.add_edge(mon_site_id, mon_site_state_id)

                if bond_num is None:
                    bond_edges[NO_BOND].append(mon_site_id)
                elif isinstance(bond_num, int):
                    bond_edges[bond_num].append(mon_site_id)

        # Unbound edges
        unbound_sites = bond_edges.pop(NO_BOND, None)
        if unbound_sites is not None:
            no_bond_id = next(node_count)
            g.add_node(no_bond_id, id=NO_BOND)
            for unbound_site in unbound_sites:
                g.add_edge(unbound_site, no_bond_id)

        # Add bond edges
        for site_nodes in bond_edges.values():
            if len(site_nodes) == 1:
                # Treat dangling bond as WILD
                any_bond_tester_id = next(node_count)
                g.add_node(any_bond_tester_id, id=any_bond_tester)
                g.add_edge(site_nodes[0], any_bond_tester_id)
            for n1, n2 in itertools.combinations(site_nodes, 2):
                g.add_edge(n1, n2)

        # Remove the species compartment if all monomer nodes have a
        # compartment
        if species_cpt_node_id is not None and \
                        g.degree(species_cpt_node_id) == 0:
            g.remove_node(species_cpt_node_id)

        self._graph = g
        return self._graph

    def is_equivalent_to(self, other):
        """
        Test a concrete ComplexPattern for equality with another.

        Use of this method on non-concrete ComplexPatterns was previously
        allowed, but is now deprecated.
        """
        from pysb.pattern import match_complex_pattern
        # Didn't implement __eq__ to avoid confusion with __ne__ operator used
        # for Rule building

        # Check both patterns are concrete
        if not self.is_concrete() or not other.is_concrete():
            warnings.warn("is_equivalent_to() will only work with concrete "
                          "patterns in a future version", DeprecationWarning)

        return match_complex_pattern(self, other, exact=True)

    def matches(self, other):
        """
        Compare another ComplexPattern against this one

        Parameters
        ----------
        other: ComplexPattern
            A ComplexPattern to match against self

        Returns
        -------
        bool
            True if other matches self; False otherwise.

        """
        if not self.is_concrete():
            raise ValueError('matches() requires self to be a concrete '
                             'pattern')
        from pysb.pattern import match_complex_pattern
        return match_complex_pattern(other, self, exact=False)

    def copy(self):
        """
        Implement our own brand of shallow copy.

        The new object will have references to the original compartment, and
        copies of the monomer_patterns.
        """
        return ComplexPattern([mp() for mp in self.monomer_patterns], self.compartment, self.match_once)

    def __call__(self, conditions=None, **kwargs):
        """Build a new ComplexPattern with updated site conditions."""

        kwargs = extract_site_conditions(conditions, **kwargs)

        # Ensure we don't have more than one of any Monomer in our patterns.
        mon_counts = collections.Counter(mp.monomer.name for mp in
                                         self.monomer_patterns)
        dup_monomers = [mon for mon, count in mon_counts.items() if count > 1]
        if dup_monomers:
            raise DuplicateMonomerError("ComplexPattern has duplicate "
                                        "Monomers: " + str(dup_monomers))

        # Ensure all specified sites are present in some Monomer.
        self_site_groups = (mp.monomer.sites for mp in self.monomer_patterns)
        self_sites = list(itertools.chain(*self_site_groups))
        unknown_sites = set(kwargs).difference(self_sites)
        if unknown_sites:
            raise UnknownSiteError("Unknown sites in argument list: " +
                                   ", ".join(unknown_sites))

        # Ensure no specified site is present in multiple Monomers.
        used_sites = [s for s in self_sites if s in kwargs]
        sgroups = itertools.groupby(sorted(used_sites))
        scounts = [(name, sum(1 for s in sites)) for name, sites in sgroups]
        dup_sites = [name for name, count in scounts if count > 1]
        if dup_sites:
            raise DuplicateSiteError("ComplexPattern has duplicate sites: " +
                                     str(dup_sites))

        # Copy self so we can modify it in place before returning it.
        cp = self.copy()
        # Build map from site name to MonomerPattern.
        site_map = {}
        for mp in cp.monomer_patterns:
            site_map.update(dict.fromkeys(mp.monomer.sites, mp))
        # Apply kwargs to our ComplexPatterns.
        for site, condition in kwargs.items():
            site_map[site].site_conditions[site] = condition
        return cp

    def __add__(self, other):
        if isinstance(other, ComplexPattern):
            return ReactionPattern([self, other])
        elif isinstance(other, MonomerPattern):
            return ReactionPattern([self, ComplexPattern([other], None)])
        elif other is None:
            rp = as_reaction_pattern(self)
            rp.complex_patterns.append(None)
            return rp
        else:
            return NotImplemented

    def __radd__(self, other):
        if other is None:
            rp = as_reaction_pattern(self)
            rp.complex_patterns = [None] + rp.complex_patterns
            return rp
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, MonomerPattern):
            return ComplexPattern(self.monomer_patterns + [other], self.compartment, self.match_once)
        elif isinstance(other, ComplexPattern):
            if self.compartment is not other.compartment:
                raise ValueError("merged ComplexPatterns must specify the same compartment")
            elif self.match_once != other.match_once:
                raise ValueError("merged ComplexPatterns must have the same value of match_once")
            return ComplexPattern(self.monomer_patterns + other.monomer_patterns, self.compartment, self.match_once)
        else:
            return NotImplemented

    def __rmod__(self, other):
        if isinstance(other, MonomerPattern):
            return ComplexPattern([other] + self.monomer_patterns, self.compartment, self.match_once)
        else:
            return NotImplemented

    def __rshift__(self, other):
        return build_rule_expression(self, other, False)

    def __rrshift__(self, other):
        return build_rule_expression(other, self, False)

    def __or__(self, other):
        return build_rule_expression(self, other, True)

    def __ne__(self, other):
        warnings.warn("'<>' for reversible rules will be removed in a future "
                      "version of PySB. Use '|' instead.",
                      PendingDeprecationWarning,
                      stacklevel=2)
        return self.__or__(other)

    def __pow__(self, other):
        if isinstance(other, Compartment):
            if self.compartment is not None:
                raise CompartmentAlreadySpecifiedError()
            cp_new = self.copy()
            cp_new.compartment = other
            return cp_new
        else:
            return NotImplemented

    def __repr__(self):
        ret = ' % '.join([repr(p) for p in self.monomer_patterns])
        if self.compartment is not None:
            ret = '(%s) ** %s' % (ret, self.compartment.name)
        if self.match_once:
            ret = 'MatchOnce(%s)' % ret
        return ret


class ReactionPattern(object):

    """
    A pattern for the entire product or reactant side of a rule.

    Essentially a thin wrapper around a list of ComplexPatterns. In BNG terms, a
    list of complex patterns combined with the '+' operator.

    Parameters
    ----------
    complex_patterns : list of ComplexPatterns
        ComplexPatterns that make up the reaction pattern.

    Attributes
    ----------
    Identical to Parameters (see above).

    """

    def __init__(self, complex_patterns):
        self.complex_patterns = complex_patterns
        from pysb.pattern import check_dangling_bonds
        check_dangling_bonds(self)

    def __add__(self, other):
        if isinstance(other, MonomerPattern):
            return ReactionPattern(self.complex_patterns + [ComplexPattern([other], None)])
        elif isinstance(other, ComplexPattern):
            return ReactionPattern(self.complex_patterns + [other])
        elif other is None:
            self.complex_patterns.append(None)
            return self
        else:
            return NotImplemented

    def __radd__(self, other):
        if other is None:
            self.complex_patterns = [None] + self.complex_patterns
            return self
        else:
            return NotImplemented

    def __rshift__(self, other):
        """Irreversible reaction"""
        return build_rule_expression(self, other, False)

    def __rrshift__(self, other):
        return build_rule_expression(other, self, False)

    def __or__(self, other):
        return build_rule_expression(self, other, True)

    def __ne__(self, other):
        warnings.warn("'<>' for reversible rules will be removed in a future "
                      "version of PySB. Use '|' instead.",
                      PendingDeprecationWarning,
                      stacklevel=2)
        return self.__or__(other)

    def __repr__(self):
        if len(self.complex_patterns):
            return ' + '.join([repr(p) for p in self.complex_patterns])
        else:
            return 'None'

    def matches(self, other):
        """
        Match the 'other' ReactionPattern against this one

        See :func:`pysb.pattern.match_reaction_pattern` for details
        """
        from pysb.pattern import match_reaction_pattern
        return match_reaction_pattern(other, self)


class RuleExpression(object):

    """
    A container for the reactant and product patterns of a rule expression.

    Contains one ReactionPattern for each of reactants and products, and a bool
    indicating reversibility. This is a temporary object used to implement
    syntactic sugar through operator overloading. The Rule constructor takes an
    instance of this class as its first argument, but simply extracts its fields
    and discards the object itself.

    Parameters
    ----------
    reactant_pattern, product_pattern : ReactionPattern
        The reactants and products of the rule.
    is_reversible : bool
        If True, the reaction is reversible. If False, it's irreversible.

    Attributes
    ----------
    Identical to Parameters (see above).

    """

    def __init__(self, reactant_pattern, product_pattern, is_reversible):
        self.reactant_pattern = reactant_pattern
        self.product_pattern = product_pattern
        self.is_reversible = is_reversible

    def __repr__(self):
        operator = '|' if self.is_reversible else '>>'
        return '%s %s %s' % (repr(self.reactant_pattern), operator,
                             repr(self.product_pattern))


def as_complex_pattern(v):
    """Internal helper to 'upgrade' a MonomerPattern to a ComplexPattern."""
    if isinstance(v, ComplexPattern):
        return v
    elif isinstance(v, Monomer):
        return ComplexPattern([v()], None)
    elif isinstance(v, MonomerPattern):
        return ComplexPattern([v], None)
    else:
        raise InvalidComplexPatternException


def as_reaction_pattern(v):
    """Internal helper to 'upgrade' a Complex- or MonomerPattern or None to a
    complete ReactionPattern."""
    if isinstance(v, ReactionPattern):
        return v
    elif v is None:
        return ReactionPattern([])
    else:
        try:
            return ReactionPattern([as_complex_pattern(v)])
        except InvalidComplexPatternException:
            raise InvalidReactionPatternException


def build_rule_expression(reactant, product, is_reversible):
    """Internal helper for operators which return a RuleExpression."""
    # Make sure the types of both reactant and product are acceptable.
    try:
        reactant = as_reaction_pattern(reactant)
        product = as_reaction_pattern(product)
    except InvalidReactionPatternException:
        return NotImplemented
    # Synthesis/degradation rules cannot be reversible.
    if (reactant is None or product is None) and is_reversible:
        raise InvalidReversibleSynthesisDegradationRule
    return RuleExpression(reactant, product, is_reversible)


class Parameter(Component, sympy.Symbol):

    """
    Model component representing a named constant floating point number.

    Parameters are used as reaction rate constants, compartment volumes and
    initial (boundary) conditions for species.

    Parameters
    ----------
    value : number, optional
        The numerical value of the parameter. Defaults to 0.0 if not specified.
        The provided value is converted to a float before being stored, so any
        value that cannot be coerced to a float will trigger an exception.

    Attributes
    ----------
    Identical to Parameters (see above).

    """

    def __new__(cls, name, value=0.0, _export=True):
        return super(sympy.Symbol, cls).__new__(cls, name)

    def __getnewargs__(self):
        return (self.name, self.value, False)

    def __init__(self, name, value=0.0, _export=True):
        Component.__init__(self, name, _export)
        self.value = float(value)
    
    def get_value(self):
        return self.value
    
    # This is needed to make sympy's evalf machinery treat this class like a
    # Symbol.
    @property
    def func(self):
        return sympy.Symbol

    def __repr__(self):
        return  '%s(%s, %s)' % (self.__class__.__name__, repr(self.name), repr(self.value))

    def __str__(self):
        return  repr(self)



class Compartment(Component):

    """
    Model component representing a bounded reaction volume.

    Parameters
    ----------
    parent : Compartment, optional
        Compartment which contains this one. If not specified, this will be the
        outermost compartment and its parent will be set to None.
    dimension : integer, optional
        The number of spatial dimensions in the compartment, either 2 (i.e. a
        membrane) or 3 (a volume).
    size : Parameter, optional
        A parameter object whose value defines the volume or area of the
        compartment. If not specified, the size will be fixed at 1.0.

    Attributes
    ----------
    Identical to Parameters (see above).

    Notes
    -----
    The compartments of a model must form a tree via their `parent` attributes
    with a three-dimensional (volume) compartment at the root. A volume
    compartment may have any number of two-dimensional (membrane) compartments
    as its children, but never another volume compartment. A membrane
    compartment may have a single volume compartment as its child, but nothing
    else.

    Examples
    --------
    Compartment('cytosol', dimension=3, size=cyto_vol, parent=ec_membrane)

    """

    def __init__(self, name, parent=None, dimension=3, size=None, _export=True):
        Component.__init__(self, name, _export)
        if parent != None and isinstance(parent, Compartment) == False:
            raise Exception("parent must be a predefined Compartment or None")
        #FIXME: check for only ONE "None" parent? i.e. only one compartment can have a parent None?
        if size is not None and not isinstance(size, Parameter):
            raise Exception("size must be a parameter (or omitted)")
        self.parent = parent
        self.dimension = dimension
        self.size = size

    def __repr__(self):
        return '%s(name=%s, parent=%s, dimension=%s, size=%s)' % (
            self.__class__.__name__,
            repr(self.name),
            'None' if self.parent is None else self.parent.name,
            repr(self.dimension),
            'None' if self.size is None else self.size.name
        )


class Rule(Component):

    """
    Model component representing a reaction rule.

    Parameters
    ----------
    rule_expression : RuleExpression
        RuleExpression containing the essence of the rule (reactants, products,
        reversibility).
    rate_forward : Parameter
        Forward reaction rate constant.
    rate_reverse : Parameter, optional
        Reverse reaction rate constant (only required for reversible rules).
    delete_molecules : bool, optional
        If True, deleting a Monomer from a species is allowed to fragment the
        species into multiple pieces (if the deleted Monomer was the sole link
        between those pieces). If False (default) then fragmentation is
        disallowed and the rule will not match a reactant species if applying
        the rule would fragment a species.
    move_connected : bool, optional
        If True, a rule that transports a Monomer between compartments will
        co-transport anything connected to that Monomer by a path in the same
        compartment. If False (default), connected Monomers will remain where
        they were.

    Attributes
    ----------

    Identical to Parameters (see above), plus the component elements of
    `rule_expression`: reactant_pattern, product_pattern and is_reversible.

    """

    def __init__(self, name, rule_expression, rate_forward, rate_reverse=None,
                 delete_molecules=False, move_connected=False,
                 _export=True):
        Component.__init__(self, name, _export)
        if not isinstance(rule_expression, RuleExpression):
            raise Exception("rule_expression is not a RuleExpression object")
        validate_expr(rate_forward, "forward rate")
        if rule_expression.is_reversible:
            validate_expr(rate_reverse, "reverse rate")
        self.rule_expression = rule_expression
        self.reactant_pattern = rule_expression.reactant_pattern
        self.product_pattern = rule_expression.product_pattern
        self.is_reversible = rule_expression.is_reversible
        self.rate_forward = rate_forward
        self.rate_reverse = rate_reverse
        self.delete_molecules = delete_molecules
        self.move_connected = move_connected
        # TODO: ensure all numbered sites are referenced exactly twice within each of reactants and products

    def is_synth(self):
        """Return a bool indicating whether this is a synthesis rule."""
        return len(self.reactant_pattern.complex_patterns) == 0 or \
            (self.is_reversible and
             len(self.product_pattern.complex_patterns) == 0)

    def is_deg(self):
        """Return a bool indicating whether this is a degradation rule."""
        return len(self.product_pattern.complex_patterns) == 0 or \
            (self.is_reversible and
             len(self.reactant_pattern.complex_patterns) == 0)

    def __repr__(self):
        ret = '%s(%s, %s, %s' % \
            (self.__class__.__name__, repr(self.name),
             repr(self.rule_expression), self.rate_forward.name)
        if self.is_reversible:
            ret += ', %s' % self.rate_reverse.name
        if self.delete_molecules:
            ret += ', delete_molecules=True'
        if self.move_connected:
            ret += ', move_connected=True'
        ret += ')'
        return ret



def validate_expr(obj, description):
    """Raises an exception if the argument is not an expression."""
    if not isinstance(obj, (Parameter, Expression)):
        description_upperfirst = description[0].upper() + description[1:]
        msg = "%s must be a Parameter or Expression" % description_upperfirst
        raise ExpressionError(msg)

def validate_const_expr(obj, description):
    """Raises an exception if the argument is not a constant expression."""
    validate_expr(obj, description)
    if isinstance(obj, Expression) and not obj.is_constant_expression():
        description_upperfirst = description[0].upper() + description[1:]
        msg = ("%s must be a Parameter or constant Expression" %
               description_upperfirst)
        raise ConstantExpressionError(msg)



class Observable(Component, sympy.Symbol):

    """
    Model component representing a linear combination of species.

    Observables are useful in correlating model simulation results with
    experimental measurements. For example, an observable for "A()" will report
    on the total number of copies of Monomer A, regardless of what it's bound to
    or the state of its sites. "A(y='P')" would report on all instances of A
    with site 'y' in state 'P'.

    Parameters
    ----------
    reaction_pattern : ReactionPattern
        The list of ComplexPatterns to match.
    match : 'species' or 'molecules'
        Whether to match entire species ('species') or individual fragments
        ('molecules'). Default is 'molecules'.

    Attributes
    ----------
    reaction_pattern : ReactionPattern
        See Parameters.
    match : 'species' or 'molecules'
        See Parameters.
    species : list of integers
        List of species indexes for species matching the pattern.
    coefficients : list of integers
        List of coefficients by which each species amount is to be multiplied to
        correct for multiple pattern matches within a species.

    Notes
    -----
    ReactionPattern is used here as a container for a list of ComplexPatterns,
    solely so users could utilize the ComplexPattern '+' operator overload as
    syntactic sugar. There are no actual "reaction" semantics in this context.

    """

    def __new__(cls, name, reaction_pattern, match='molecules', _export=True):
        return super(sympy.Symbol, cls).__new__(cls, name)

    def __getnewargs__(self):
        return (self.name, self.reaction_pattern, self.match, False)

    def __init__(self, name, reaction_pattern, match='molecules', _export=True):
        try:
            reaction_pattern = as_reaction_pattern(reaction_pattern)
        except InvalidReactionPatternException as e:
            raise type(e)("Observable pattern does not look like a ReactionPattern")
        if match not in ('molecules', 'species'):
            raise ValueError("Match must be 'molecules' or 'species'")
        Component.__init__(self, name, _export)
        self.reaction_pattern = reaction_pattern
        self.match = match
        self.species = []
        self.coefficients = []

    # This is needed to make sympy's evalf machinery treat this class like a
    # Symbol.
    @property
    def func(self):
        return sympy.Symbol

    def expand_obs(self):
        """ Expand observables in terms of species and coefficients """
        return sympy.Add(*[a * b for a, b in zip(
            self.coefficients,
            sympy.symbols(','.join('__s%d' % sp_id for sp_id in
                                   self.species) + ',')
        )])

    def __repr__(self):
        ret = '%s(%s, %s' % (self.__class__.__name__, repr(self.name),
                              repr(self.reaction_pattern))
        if self.match != 'molecules':
            ret += ', match=%s' % repr(self.match)
        ret += ')'
        return ret

    def __str__(self):
        return repr(self)


class Expression(Component, sympy.Symbol):

    """
    Model component representing a symbolic expression of other variables.

    Parameters
    ----------
    expr : sympy.Expr
        Symbolic expression.

    Attributes
    ----------
    expr : sympy.Expr
        See Parameters.

    """

    def __new__(cls, name, expr, _export=True):
        return super(sympy.Symbol, cls).__new__(cls, name)

    def __getnewargs__(self):
        return (self.name, self.expr, False)

    def __init__(self, name, expr, _export=True):
        Component.__init__(self, name, _export)
        self.expr = expr

    def expand_expr(self, expand_observables=False):
        """Return expr rewritten in terms of terminal symbols only."""
        subs = []
        for a in self.expr.atoms():
            if isinstance(a, Expression):
                subs.append((a, a.expand_expr(
                    expand_observables=expand_observables)))
            elif expand_observables and isinstance(a, Observable):
                subs.append((a, a.expand_obs()))
        return self.expr.subs(subs)

    def is_constant_expression(self):
        """Return True if all terminal symbols are Parameters or numbers."""
        return all(isinstance(a, Parameter) or
                   (isinstance(a, Expression) and a.is_constant_expression()) or
                   isinstance(a, sympy.Number)
                   for a in self.expr.atoms())

    def get_value(self):
        return self.expr.evalf()

    # This is needed to make sympy's evalf machinery treat this class like a
    # Symbol.
    @property
    def func(self):
        return sympy.Symbol

    def __repr__(self):
        ret = '%s(%s, %s)' % (self.__class__.__name__, repr(self.name),
                              repr(self.expr))
        return ret

    def __str__(self):
        return repr(self)


class Model(object):

    """
    A rule-based model containing monomers, rules, compartments and parameters.

    Parameters
    ----------
    name : string, optional
        Name of the model. If not specified, will be set to the name of the file
        from which the constructor was called (with the .py extension stripped).
    base : Model, optional
        If specified, the model will begin as a copy of `base`. This can be used
        to achieve a simple sort of model extension and enhancement.

    Attributes
    ----------
    name : string
        Name of the model. See Parameter section above.
    base : Model or None
        See Parameter section above.
    monomers, compartments, parameters, rules, observables : ComponentSet
        The Component objects which make up the model.
    initial_conditions : list of tuple of (ComplexPattern, Parameter)
        Specifies which species are present in the model's starting
        state (t=0) and how much there is of each one.  The
        ComplexPattern defines the species identity, and it must be
        concrete (see ComplexPattern.is_concrete).  The
        Parameter defines the amount or concentration of the species.
    species : list of ComplexPattern
        List of all complexes which can be produced by the model, starting from
        the initial conditions and successively applying the rules. Each 
        ComplexPattern is concrete.
    reactions : list of dict
        Structures describing each possible unidirectional reaction that can be
        produced by the model. Each structure stores the name of the rule that
        generated the reaction ('rule'), the mathematical expression for the
        rate of the reaction ('rate'), tuples of species indexes for the
        reactants and products ('reactants', 'products'), and a bool indicating
        whether the reaction is the reverse component of a bidirectional
        reaction ('reverse').
    reactions_bidirectional : list of dict
        Similar to `reactions` but with only one entry for each bidirectional
        reaction. The fields are identical except 'reverse' is replaced by
        'reversible', a bool indicating whether the reaction is reversible. The
        'rate' is the forward rate minus the reverse rate.
    annotations : list of Annotation
        Structured annotations of model components. See the Annotation class for
        details.

    """

    _component_types = (Monomer, Compartment, Parameter, Rule, Observable,
                        Expression)

    def __init__(self, name=None, base=None, _export=True):
        self.name = name
        self.base = base
        self._export = _export
        self.monomers = ComponentSet()
        self.compartments = ComponentSet()
        self.parameters = ComponentSet()
        self.rules = ComponentSet()
        self.observables = ComponentSet()
        self.expressions = ComponentSet()
        self.initial_conditions = []
        self.annotations = []
        self._odes = OdeView(self)
        self.reset_equations()
        #####
        self.diffusivities = []
        #####
        if self._export:
            SelfExporter.export(self)
        if self.base is not None:
            if not isinstance(self.base, Model):
                raise ValueError("base must be a Model")
            model_copy = copy.deepcopy(self.base)
            for component in model_copy.all_components():
                self.add_component(component)
                component._do_export()
            self.initial_conditions = model_copy.initial_conditions

    def __getstate__(self):
        state = self.__dict__.copy()
        # The stoichiometry matrix, as a numpy array, is problematic to pickle
        # in a cross-Python-version-compatible way. Since it's regenerated on
        # demand anyway, we can just clear it here.
        state['_stoichiometry_matrix'] = None
        return state

    def __setstate__(self, state):
        # restore the 'model' weakrefs on all components
        self.__dict__.update(state)
        for c in self.all_components():
            c.model = weakref.ref(self)

    def reload(self):
        """
        Reload a model after its source files have been edited.

        This method does not yet reload the model contents in-place, rather it
        returns a new model object. Thus the correct usage is ``model =
        model.reload()``.

        If the model script imports any modules, these will not be reloaded. Use
        python's reload() function to reload them.

        """
        # forcibly removes the .pyc file and reloads the model module
        model_pyc = SelfExporter.target_module.__file__
        if model_pyc[-3:] == '.py':
            model_pyc += 'c'
        try:
            os.unlink(model_pyc)
        except OSError as e:
            # ignore "no such file" errors, re-raise the rest
            if e.errno != errno.ENOENT:
                raise
        try:
            reload(SelfExporter.target_module)
        except SystemError as e:
            # This one specific SystemError occurs when using ipython to 'run' a model .py file
            # directly, then reload()ing the model, which makes no sense anyway. (just re-run it)
            if e.args == ('nameless module',):
                raise Exception('Cannot reload a model which was executed directly in an interactive'
                                'session. Please import the model file as a module instead.')
            else:
                raise
        # return self for "model = model.reload()" idiom, until a better solution can be found
        return SelfExporter.default_model

    def all_component_sets(self):
        """Return a list of all ComponentSet objects."""
        set_names = [t.__name__.lower() + 's' for t in Model._component_types]
        sets = [getattr(self, name) for name in set_names]
        return sets

    def all_components(self):
        """Return a ComponentSet containing all components in the model."""
        cset_all = ComponentSet()
        for cset in self.all_component_sets():
            cset_all |= cset
        return cset_all

    @property
    def components(self):
        return self.all_components()

    def parameters_rules(self):
        """Return a ComponentSet of the parameters used in rules."""
        # rate_reverse is None for irreversible rules, so we'll need to filter those out
        cset = ComponentSet(p for r in self.rules for p in (r.rate_forward, r.rate_reverse)
                            if p is not None)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_initial_conditions(self):
        """Return a ComponentSet of initial condition parameters."""
        cset = ComponentSet(ic[1] for ic in self.initial_conditions)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_compartments(self):
        """Return a ComponentSet of compartment size parameters."""
        cset = ComponentSet(c.size for c in self.compartments)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_unused(self):
        """Return a ComponentSet of unused parameters."""
        cset_used = self.parameters_rules() | self.parameters_initial_conditions() | self.parameters_compartments()
        return self.parameters - cset_used

#     def expressions_constant(self):
#         """Return a ComponentSet of constant expressions."""
#         cset = ComponentSet(e for e in self.expressions
#                             if all(isinstance(a, (Parameter, sympy.Number))
#                                    for a in e.expand_expr().atoms()))
#         return cset
    
    def expressions_constant(self):
        """Return a ComponentSet of constant expressions."""
        cset = ComponentSet(e for e in self.expressions
                            if e.is_constant_expression())
        return cset

    def expressions_dynamic(self):
        """Return a ComponentSet of non-constant expressions."""
        return self.expressions - self.expressions_constant()

    @property
    def odes(self):
        """Return sympy Expressions for the time derivative of each species."""
        return self._odes

    @property
    def stoichiometry_matrix(self):
        """Return the stoichiometry matrix for the reaction network."""
        if self._stoichiometry_matrix is None:
            shape = (len(self.species), len(self.reactions))
            sm = scipy.sparse.lil_matrix(shape, dtype='int')
            for i, reaction in enumerate(self.reactions):
                for r in reaction['reactants']:
                    sm[r, i] -= 1
                for p in reaction['products']:
                    sm[p, i] += 1
            self._stoichiometry_matrix = sm.tocsr()
        return self._stoichiometry_matrix

    def add_component(self, other):
        """Add a component to the model."""
        # We have a container for each type of component. This code determines
        # the right one based on the class of the object being added.
        for t, cset in zip(Model._component_types, self.all_component_sets()):
            if isinstance(other, t):
                cset.add(other)
                other.model = weakref.ref(self)
                break
        else:
            raise Exception("Tried to add component of unknown type '%s' to "
                            "model" % type(other))

    def add_annotation(self, annotation):
        """Add an annotation to the model."""
        self.annotations.append(annotation)

    def get_annotations(self, subject):
        """Return all annotations for the given subject."""
        annotations = []
        for a in self.annotations:
            if a.subject is subject:
                annotations.append(a)
        return annotations

    def _rename_component(self, component, new_name):
        """
        Change a component's name.

        This has to be done through the Model because the ComponentSet needs to
        be updated as well as the component's `name` field.

        """
        for cset in self.all_component_sets():
            if component in cset:
                cset.rename(component, new_name)

    def _validate_initial_condition_pattern(self, pattern):
        """
        Make sure a pattern is valid for an initial condition.

        Patterns must satisfy all of the following:
        * Able to be cast as a ComplexPattern
        * Concrete (see ComplexPattern.is_concrete)
        * Distinct from any existing initial condition pattern
        * match_once is False (nonsensical in this context)

        Parameters
        ----------
        pattern : MonomerPattern or ComplexPattern
            Pattern to validate

        Returns
        -------
        The validated pattern, upgraded to a ComplexPattern.

        """
        try:
            complex_pattern = as_complex_pattern(pattern)
        except InvalidComplexPatternException as e:
            raise InvalidInitialConditionError("Not a ComplexPattern")
        if not complex_pattern.is_concrete():
            raise InvalidInitialConditionError("Pattern not concrete")
        if any(complex_pattern.is_equivalent_to(other_cp)
               for other_cp, value in self.initial_conditions):
            # FIXME until we get proper canonicalization this could produce
            # false negatives
            raise InvalidInitialConditionError("Duplicate species")
        if complex_pattern.match_once:
            raise InvalidInitialConditionError("MatchOnce not allowed here")
        return complex_pattern

    def initial(self, pattern, value):
        """
        Add an initial condition.

        An initial condition is made up of a species and its amount or
        concentration.

        Parameters
        ----------
        pattern : ComplexPattern
            A concrete pattern defining the species to initialize.
        value : Parameter
            Amount of the species the model will start with.

        """
        complex_pattern = self._validate_initial_condition_pattern(pattern)
        validate_const_expr(value, "initial condition value")
        self.initial_conditions.append( (complex_pattern, value) )

    def update_initial_condition_pattern(self, before_pattern, after_pattern):
        """
        Update the pattern associated with an initial condition.

        Leaves the Parameter object associated with the initial condition
        unchanged while modifying the pattern associated with that condition.
        For example this is useful for changing the state of a site on a
        monomer or complex associated with an initial condition without having
        to create an independent initial condition, and parameter, associated
        with that alternative state.

        Parameters
        ----------
        before_pattern : ComplexPattern
            The concrete pattern specifying the (already existing) initial
            condition. If the model does not contain an initial condition
            for the pattern, a ValueError is raised.
        after_pattern : ComplexPattern
            The concrete pattern specifying the new pattern to use to replace
            before_pattern.
        """

        # Get the initial condition index
        ic_index_list = [i for i, ic in enumerate(self.initial_conditions)
                   if ic[0].is_equivalent_to(as_complex_pattern(before_pattern))]

        # If the initial condition to replace is not found, raise an error
        if not ic_index_list:
            raise ValueError("No initial condition found for pattern %s" %
                             before_pattern)

        # If more than one matching initial condition is found, raise an
        # error (this should never happen, because duplicate initial conditions
        # are not allowed to be created)
        assert len(ic_index_list) == 1
        ic_index = ic_index_list[0]

        # Make sure the new initial condition pattern is valid
        after_pattern = self._validate_initial_condition_pattern(after_pattern)

        # Since everything checks out, replace the old initial condition
        # pattern with the new one.  Because initial_conditions are tuples (and
        # hence immutable), we cannot simply replace the pattern; instead we
        # must delete the old one and add the new one.
        # We retain the old parameter object:
        p = self.initial_conditions[ic_index][1]
        del self.initial_conditions[ic_index]
        self.initial_conditions.append( (after_pattern, p) )

    def get_species_index(self, complex_pattern):
        """
        Return the index of a species.

        Parameters
        ----------
        complex_pattern : ComplexPattern
            A concrete pattern specifying the species to find.

        """
        # FIXME I don't even want to think about the inefficiency of this, but at least it works
        try:
            return next((i for i, s_cp in enumerate(self.species) if s_cp.is_equivalent_to(complex_pattern)))
        except StopIteration:
            return None

    def has_synth_deg(self):
        """Return true if model uses synthesis or degradation reactions."""
        return any(r.is_synth() or r.is_deg() for r in self.rules)

    def enable_synth_deg(self):
        """Add components needed to support synthesis and degradation rules."""
        warnings.warn('This function is no longer needed, and no longer has '
                      'any effect.', DeprecationWarning)

    def reset_equations(self):
        """Clear out fields generated by bng.generate_equations or the like."""
        self.species = []
        self.reactions = []
        self.reactions_bidirectional = []
        self._stoichiometry_matrix = None
        for obs in self.observables:
            obs.species = []
            obs.coefficients = []

    def __repr__(self):
        return ("<%s '%s' (monomers: %d, rules: %d, parameters: %d, "
                "expressions: %d, compartments: %d) at 0x%x>" %
                (self.__class__.__name__, self.name,
                 len(self.monomers), len(self.rules), len(self.parameters),
                 len(self.expressions), len(self.compartments), id(self)))



class InvalidComplexPatternException(Exception):
    """Expression can not be cast as a ComplexPattern."""
    pass

class InvalidReactionPatternException(Exception):
    """Expression can not be cast as a ReactionPattern."""
    pass

class InvalidReversibleSynthesisDegradationRule(Exception):
    """Synthesis or degradation rule defined as reversible."""
    def __init__(self):
        Exception.__init__(self, "Synthesis and degradation rules may not be"
                           "reversible.")

class ExpressionError(ValueError):
    """Expected an Expression but got something else."""
    pass

class ConstantExpressionError(ValueError):
    """Expected a constant Expression but got something else."""
    pass

class ModelExistsWarning(UserWarning):
    """A second model was declared in a module that already contains one."""
    pass

class SymbolExistsWarning(UserWarning):
    """A component declaration or rename overwrote an existing symbol."""
    pass

class InvalidComponentNameError(ValueError):
    """Inappropriate component name."""
    def __init__(self, name):
        ValueError.__init__(self, "Not a valid component name: '%s'" % name)

class InvalidInitialConditionError(ValueError):
    """Invalid initial condition pattern."""

class DuplicateMonomerError(ValueError):
    pass

class DuplicateSiteError(ValueError):
    pass

class UnknownSiteError(ValueError):
    pass

class CompartmentAlreadySpecifiedError(ValueError):
    pass


class ComponentSet(collections.Set, collections.Mapping, collections.Sequence):
    """
    An add-and-read-only container for storing model Components.

    It behaves mostly like an ordered set, but components can also be retrieved
    by name *or* index by using the [] operator (like a combination of a dict
    and a list). Components cannot be removed or replaced, but they can be
    renamed. Iteration returns the component objects.

    Parameters
    ----------
    iterable : iterable of Components, optional
        Initial contents of the set.

    """

    # The implementation is based on a list instead of a linked list (as
    # OrderedSet is), since we only allow add and retrieve, not delete.

    def __init__(self, iterable=None):
        self._elements = []
        self._map = {}
        self._index_map = {}
        if iterable is not None:
            for value in iterable:
                self.add(value)

    def __iter__(self):
        return iter(self._elements)

    def __contains__(self, c):
        if not isinstance(c, Component):
            raise TypeError("Can only work with Components, got a %s" % type(c))
        return c.name in self._map and self[c.name] is c

    def __len__(self):
        return len(self._elements)

    def add(self, c):
        if c not in self:
            if c.name in self._map:
                raise ComponentDuplicateNameError(
                    "Tried to add a component with a duplicate name: %s"
                    % c.name)
            self._elements.append(c)
            self._map[c.name] = c
            self._index_map[c.name] = len(self._elements) - 1

    def __getitem__(self, key):
        # Must support both Sequence and Mapping behavior. This means
        # stringified integer Mapping keys (like "0") are forbidden, but since
        # all Component names must be valid Python identifiers, integers are
        # ruled out anyway.
        if isinstance(key, (int, long, slice)):
            return self._elements[key]
        else:
            return self._map[key]

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError("Model has no component '%s'" % name)

    def __setstate__(self, state):
        self.__dict__ = state

    def get(self, key, default=None):
        if isinstance(key, (int, long)):
            raise ValueError("get is undefined for integer arguments, use []"
                             "instead")
        try:
            return self[key]
        except KeyError:
            return default

    def iterkeys(self):
        for c in self:
            yield c.name

    def itervalues(self):
        return self.__iter__()

    def iteritems(self):
        for c in self:
            yield (c.name, c)

    def keys(self):
        return [c.name for c in self]

    def values(self):
        return [c for c in self]

    def items(self):
        return list(zip(self.keys(), self))

    def index(self, c):
        # We can implement this in O(1) ourselves, whereas the Sequence mixin
        # implements it in O(n).
        if not c in self:
            raise ValueError("%s is not in ComponentSet" % c)
        return self._index_map[c.name]

    def __and__(self, other):
        # We reimplement this because collections.Set's __and__ mixin iterates
        # over other, not self. That implementation ends up retaining the
        # ordering of other, but we'd like to keep the ordering of self instead.
        # We require other to be a ComponentSet too so we know it will support
        # "in" efficiently.
        if not isinstance(other, ComponentSet):
            return collections.Set.__and__(self, other)
        return ComponentSet(value for value in self if value in other)

    def __rand__(self, other):
        return self.__and__(other)

    def __ror__(self, other):
        return self.__or__(other)

    def __rxor__(self, other):
        return self.__xor__(other)

    def __repr__(self):
        return 'ComponentSet([\n' + \
            ''.join(' %s,\n' % repr(x) for x in self) + \
            ' ])'

    def rename(self, c, new_name):
        """Change the name of component `c` to `new_name`."""
        for m in self._map, self._index_map:
            m[new_name] = m[c.name]
            del m[c.name]


class OdeView(collections.Sequence):
    """Compatibility shim for the Model.odes property."""

    # This is necessarily coupled pretty tightly with Model. Note that we
    # faithfully emulate the detail of the original implementation in which odes
    # is an empty list before the equation generation process is run (and after
    # reset_equations is called). Now the "empty" condition is when species is
    # empty.

    def __init__(self, model):
        self.model = model

    def __getitem__(self, key):
        if not self.model.species:
            raise IndexError('list index out of range')
        if isinstance(key, slice):
            return [self[k] for k in range(*key.indices(len(self)))]
        sr = self.model.stoichiometry_matrix[key]
        terms = [sympy.Mul(self.model.reactions[i]['rate'], v, evaluate=False)
                 for i, v in zip(sr.indices, sr.data)]
        return sympy.Add(*terms, evaluate=False)

    def __len__(self):
        return len(self.model.species)


class ComponentDuplicateNameError(ValueError):
    """A component was added with the same name as an existing one."""
    pass


def extract_site_conditions(conditions=None, **kwargs):
    """Parse MonomerPattern/ComplexPattern site conditions."""
    # enforce site conditions as kwargs or a dict but not both
    if conditions and kwargs:
        raise RedundantSiteConditionsError()
    # handle normal cases
    elif conditions:
        site_conditions = conditions.copy()
    else:
        site_conditions = kwargs
    return site_conditions


class RedundantSiteConditionsError(ValueError):
    """Both conditions dict and kwargs both passed to create pattern."""
    def __init__(self):
        ValueError.__init__(
            self,
            ("Site conditions may be specified as EITHER keyword arguments "
             "OR a single dict"))


class DanglingBondError(ValueError):
    pass

# Some light infrastructure for defining symbols that act like "keywords", i.e.
# they are immutable singletons that stringify to their own name. Regular old
# classes almost fit the bill, except that their __str__ method prepends the
# complete module hierarchy to the base class name. The KeywordMeta class here
# implements an alternate __str__ method which just returns the base name.

class KeywordMeta(type):
    def __repr__(cls):
        return cls.__name__
    def __str__(cls):
        return repr(cls)


# Define Keyword class with KeywordMeta metaclass in a Python 2 and 3
# compatible way
class Keyword(KeywordMeta("KeywordMetaBase", (object, ), {})):
    pass

# The keywords.

class ANY(Keyword):
    """Site must have a bond, but identity of binding partner is irrelevant.

    Use ANY in a MonomerPattern site_conditions dict to indicate that a site
    must have a bond without specifying what the binding partner should be.

    Equivalent to the "+" bond modifier in BNG."""
    pass

class WILD(Keyword):
    """Site may be bound or unbound.

    Use WILD as part of a (state, WILD) tuple in a MonomerPattern
    site_conditions dict to indicate that a site must have the given state,
    irrespective of the presence or absence of a bond. (Specifying only the
    state implies there must not be a bond). A bare WILD in a site_conditions
    dict is also permissible, but as this has the same meaning as the much
    simpler option of leaving the given site out of the dict entirely, this
    usage is deprecated.

    Equivalent to the "?" bond modifier in BNG."""
    pass


warnings.simplefilter('always', ModelExistsWarning)
warnings.simplefilter('always', SymbolExistsWarning)



#!/usr/bin/env python
""" Run the Extrinsic Apoptosis Reaction Model (EARM) using BioNetGen's
Hybrid-Particle Population (HPP) algorithm.

NFsim provides stochastic simulation without reaction network generation, 
allowing simulation of models with large (or infinite) reaction networks by 
keeping track of species counts. However, it can fail when the number of 
instances of a species gets too large (typically >200000). HPP circumvents 
this problem by allowing the user to define species with large instance 
counts as populations rather than NFsim particles.

This example runs the EARM 1.0 model with HPP, which fails to run on NFsim 
with the default settings due to large initial concentration coutns of 
several species. By assigning population maps to these species, we can run 
the simulation.

Reference: Hogg et al. Plos Comb Biol 2014
           https://doi.org/10.1371/journal.pcbi.1003544
"""
from pysb.examples.earm_1_0 import model
from pysb.simulator import BngSimulator
from pysb.simulator.bng import PopulationMap
from pysb import Parameter
import matplotlib.pyplot as plt
import numpy as np


def plot_mean_min_max(name, title=None):
    x = np.array([tr[:][name] for tr in trajectories]).T
    if not title:
        title = name
    plt.figure(title)
    plt.plot(tout.T, x, '0.5', lw=2, alpha=0.25) # individual trajectories
    plt.plot(tout[0], x.mean(1), 'k--', lw=3, label="Mean")
    plt.plot(tout[0], x.min(1), 'b--', lw=3, label="Minimum")
    plt.plot(tout[0], x.max(1), 'r--', lw=3, label="Maximum")
    plt.legend(loc=0)
    plt.xlabel('Time')
    plt.ylabel('Population of %s' % name)

PARP, CPARP, Mito, mCytoC = [model.monomers[x] for x in
                             ['PARP', 'CPARP', 'Mito', 'mCytoC']]
klump = Parameter('klump', 10000, _export=False)
model.add_component(klump)

population_maps = [
    PopulationMap(PARP(b=None), klump),
    PopulationMap(CPARP(b=None), klump),
    PopulationMap(Mito(b=None), klump),
    PopulationMap(mCytoC(b=None), klump)
]

sim = BngSimulator(model, tspan=np.linspace(0, 20000, 101))
simres = sim.run(n_runs=20, method='nf', population_maps=population_maps)

trajectories = simres.all
tout = simres.tout

plot_mean_min_max('Bid_unbound')
plot_mean_min_max('PARP_unbound')
plot_mean_min_max('mSmac_unbound')
plot_mean_min_max('tBid_total')
plot_mean_min_max('CPARP_total')
plot_mean_min_max('cSmac_total')

plt.show()



from pysb import *
from pysb.macros import catalyze_state

Model()

Parameter('vol', 10.) # volume (arbitrary units)

Parameter('kf',   1./vol.value)
Parameter('kr',   1000)
Parameter('kcat', 100)

Monomer('E', ['s'])
Monomer('S', ['e', 'state'], {'state': ['0','1']})

catalyze_state(E(), 's', S(), 'e', 'state', '0', '1', [kf,kr,kcat])

Observable("E_free",     E(s=None))
Observable("S_free",     S(e=None, state='0'))
Observable("ES_complex", E(s=1) % S(e=1))
Observable("Product",    S(e=None, state='1'))

Parameter("Etot", 1.*vol.value)
Initial(E(s=None), Etot)

Parameter('S0', 10.*vol.value)
Initial(S(e=None, state='0'), S0)



import os
import sys
import sysconfig

_path_config = {
    'bng': {
        'name': 'BioNetGen',
        'executable': 'BNG2.pl',
        'batch_file': 'BNG2.bat',
        'env_var': 'BNGPATH',
        'search_paths': {
            'posix': ('/usr/local/share/BioNetGen', ),
            'nt': ('c:/Program Files/BioNetGen', )
        }
    },
    'kasa': {
        'name': 'KaSa (Kappa)',
        'executable': {
            'posix': 'KaSa',
            'nt': 'KaSa.exe'
        },
        'env_var': 'KAPPAPATH',
        'search_paths': {
            'posix': ('/usr/local/share/KaSa', ),
            'nt': ('c:/Program Files/KaSa', )
        }
    },
    'kasim': {
        'name': 'KaSim (Kappa)',
        'executable': {
            'posix': 'KaSim',
            'nt': 'KaSim.exe'
        },
        'env_var': 'KAPPAPATH',
        'search_paths': {
            'posix': ('/usr/local/share/KaSim',),
            'nt': ('c:/Program Files/KaSim',)
        }
    },
    'cupsoda': {
        'name': 'cupSODA',
        'executable': {
            'posix': 'cupSODA',
            'nt': 'cupsoda.exe'
        },
        'env_var': 'CUPSODAPATH',
        'search_paths': {
            'posix': ('/usr/local/share/cupSODA',),
            'nt': ('c:/Program Files/cupSODA',)
        }
    },
    'stochkit_ssa': {
        'name': 'StochKit [SSA]',
        'executable': {
            'posix': 'ssa',
            'nt': 'ssa.exe'
        },
        'batch_file': 'ssa.bat',
        'env_var': 'STOCHKITPATH',
        'search_paths': {
            'posix': ('/usr/local/share/StochKit', ),
            'nt': ('c:/Program Files/StochKit',)
        }
    },
    'stochkit_tau_leaping': {
        'name': 'StochKit [Tau Leaping]',
        'executable': {
            'posix': 'tau_leaping',
            'nt': 'tau_leaping.exe'
        },
        'batch_file': 'tau_leaping.bat',
        'env_var': 'STOCHKITPATH',
        'search_paths': {
            'posix': ('/usr/local/share/StochKit',),
            'nt': ('c:/Program Files/StochKit',)
        }
    }
}
_path_cache = {}


def get_path(prog_name):
    """
    Gets the currently active path to an external executable

    The path will be determined automatically if not set (see return value).
    To override, call :func:`set_path`.

    Parameters
    ----------
    prog_name: str
        The PySB internal program name for an executable. One of 'bng'
        (BioNetGen), 'kasa' (Kappa's KaSa) or 'kasim' (Kappa's KaSim).

    Returns
    -------
    The currently active path to an external executable. If the path hasn't
    previously been set, the relevant environment variable for that
    program's path will be searched. Failing that, a list of default paths
    for the operating system will be checked. An Exception is raised if none
    of these approaches works.
    """
    try:
        return _path_cache[prog_name]
    except KeyError:
        pass

    if prog_name not in _path_config.keys():
        raise ValueError('%s is not a known external executable' % prog_name)

    path_conf = _path_config[prog_name]

    # Try environment variable, if set
    if path_conf['env_var'] in os.environ:
        env_var_val = os.environ[path_conf['env_var']]
        try:
            _path_cache[prog_name] = _validate_path(prog_name, env_var_val)
            return _path_cache[prog_name]
        except ValueError:
            raise ValueError('Environment variable %s is set to %s, but the '
                             'program %s or its executable %s could not be '
                             'found there. Check file existence and '
                             'permissions.' % (
                                path_conf['env_var'],
                                env_var_val,
                                path_conf['name'],
                                _get_executable(prog_name)))

    # If this is anaconda, check the anaconda binary dir
    if _is_anaconda():
        try:
            _path_cache[prog_name] = _validate_path(prog_name,
                                                    _get_anaconda_bindir())
            return _path_cache[prog_name]
        except ValueError:
            pass

    # Check default paths for this operating system
    if os.name not in path_conf['search_paths'].keys():
        raise Exception('No default path is known for %s on your '
                        'operating system "%s". Set the path using the '
                        'environment variable %s or by calling the function '
                        '%s.%s()' % (path_conf['name'],
                                     os.name,
                                     path_conf['env_var'],
                                     set_path.__module__,
                                     set_path.__name__))

    search_paths = path_conf['search_paths'][os.name]
    for search_path in search_paths:
        try:
            _path_cache[prog_name] = _validate_path(prog_name, search_path)
            return _path_cache[prog_name]
        except ValueError:
            pass

    raise Exception('The program %s was not found in the default search '
                    'path(s) for your operating system:\n\n%s\n\nEither '
                    'install it to one of those paths, or set a custom path '
                    'using the environment variable %s or by calling the '
                    'function %s.%s()' % (path_conf['name'],
                                          "\n".join(search_paths),
                                          path_conf['env_var'],
                                          set_path.__module__,
                                          set_path.__name__))


def set_path(prog_name, full_path):
    """
    Sets the full path to an external executable

    Parameters
    ----------
    prog_name: str
        The internal program name for an executable. See :func:`get_path` for
        valid values.
    full_path: str
        The full path to the external executable or its enclosing directory.
        If the path is a directory, it will be searched for the executable.
        A ValueError will be raised if there's an issue with the path (not
        found, permissions etc.).
    """
    if prog_name not in _path_config.keys():
        raise ValueError('%s is not a known external executable' % prog_name)

    _path_cache[prog_name] = _validate_path(prog_name, full_path)


def _is_anaconda():
    """ Identify if this is an anaconda environment """
    return 'conda' in sys.version


def _get_anaconda_bindir():
    """ Get the binary path from python build time (for anaconda) """
    # Is this an anaconda virtual environment?
    conda_env = os.environ.get('CONDA_PREFIX', None)
    if conda_env:
        return os.path.join(conda_env, 'Scripts' if os.name == 'nt' else 'bin')

    # Otherwise, use the default anaconda bin directory
    bindir = sysconfig.get_config_var('BINDIR')
    if os.name == 'nt':
        # bindir doesn't point to scripts directory on Windows
        return os.path.join(bindir, 'Scripts')
    else:
        return bindir

def _get_batch_file(prog_name):
    try:
        return _path_config[prog_name]['batch_file']
    except KeyError:
        return None


def _get_executable(prog_name):
    executable = _path_config[prog_name]['executable']
    if isinstance(executable, str):
        return executable
    else:
        try:
            return executable[os.name]
        except KeyError:
            raise Exception('No executable for "%s" is available for your '
                            'operating system: "%s"' %
                            (_path_config[prog_name]['name'],
                             os.name))


def _validate_path(prog_name, full_path):
    if not os.access(full_path, os.F_OK):
        raise ValueError('Unable to access path %s. Check the file exists '
                         'and the current user has permission to access it.' %
                         full_path)

    if not os.path.isfile(full_path):
        # On anaconda, check batch file on Windows, if applicable
        batch_file = _get_batch_file(prog_name)
        if _is_anaconda() and os.name == 'nt' and batch_file:
            try:
                return _validate_path(prog_name,
                                      os.path.join(full_path, batch_file))
            except ValueError:
                pass

        # It's a directory, try appending the executable name
        return _validate_path(prog_name, os.path.join(full_path,
                                                      _get_executable(
                                                          prog_name)))

    if not os.access(full_path, os.X_OK):
        raise ValueError('The file %s does not have executable permissions.' %
                         full_path)

    return full_path



from pysb.testing import *
import numpy as np
from pysb import Monomer, Parameter, Initial, Observable, Rule
from pysb.simulator.bng import BngSimulator, PopulationMap
from pysb.bng import generate_equations
from pysb.examples import robertson, expression_observables, earm_1_0

_BNG_SEED = 123


class TestBngSimulator(object):
    @with_model
    def setUp(self):
        Monomer('A', ['a'])
        Monomer('B', ['b'])

        Parameter('ksynthA', 100)
        Parameter('ksynthB', 100)
        Parameter('kbindAB', 100)

        Parameter('A_init', 0)
        Parameter('B_init', 0)

        Initial(A(a=None), A_init)
        Initial(B(b=None), B_init)

        Observable("A_free", A(a=None))
        Observable("B_free", B(b=None))
        Observable("AB_complex", A(a=1) % B(b=1))

        Rule('A_synth', None >> A(a=None), ksynthA)
        Rule('B_synth', None >> B(b=None), ksynthB)
        Rule('AB_bind', A(a=None) + B(b=None) >> A(a=1) % B(b=1), kbindAB)

        self.model = model
        generate_equations(self.model)

        # Convenience shortcut for accessing model monomer objects
        self.mon = lambda m: self.model.monomers[m]

        # This timespan is chosen to be enough to trigger a Jacobian evaluation
        # on the various solvers.
        self.time = np.linspace(0, 1)
        self.sim = BngSimulator(self.model, tspan=self.time)

    def test_1_simulation(self):
        x = self.sim.run()
        assert np.allclose(x.tout, self.time)

    def test_multi_simulations(self):
        x = self.sim.run(n_runs=10)
        assert np.shape(x.observables) == (10, 50)
        # Check initials are getting correctly reset on each simulation
        assert np.allclose(x.species[-1][0, :], x.species[0][0, :])

    def test_change_parameters(self):
        x = self.sim.run(n_runs=10, param_values={'ksynthA': 200},
                         initials={self.model.species[0]: 100})
        species = np.array(x.all)
        assert species[0][0][0] == 100.

    def test_bng_pla(self):
        self.sim.run(n_runs=5, method='pla', seed=_BNG_SEED)

    def tearDown(self):
        self.model = None
        self.time = None
        self.sim = None


def test_bng_ode_with_expressions():
    model = expression_observables.model
    model.reset_equations()

    sim = BngSimulator(model, tspan=np.linspace(0, 1))
    x = sim.run(n_runs=1, method='ode')
    assert len(x.expressions) == 50
    assert len(x.observables) == 50


def test_nfsim():
    model = robertson.model
    # Reset equations from any previous network generation
    model.reset_equations()

    sim = BngSimulator(model, tspan=np.linspace(0, 1))
    x = sim.run(n_runs=1, method='nf', seed=_BNG_SEED)
    observables = np.array(x.observables)
    assert len(observables) == 50

    A = model.monomers['A']
    x = sim.run(n_runs=2, method='nf', tspan=np.linspace(0, 1),
                initials={A(): 100}, seed=_BNG_SEED)
    assert np.allclose(x.dataframe.loc[0, 0.0], [100.0, 0.0, 0.0])


def test_hpp():
    model = robertson.model
    # Reset equations from any previous network generation
    model.reset_equations()

    A = robertson.model.monomers['A']
    klump = Parameter('klump', 10000, _export=False)
    model.add_component(klump)

    population_maps = [
        PopulationMap(A(), klump)
    ]

    sim = BngSimulator(model, tspan=np.linspace(0, 1))
    x = sim.run(n_runs=1, method='nf', population_maps=population_maps,
                seed=_BNG_SEED)
    observables = np.array(x.observables)
    assert len(observables) == 50



from pysb.testing import *
import sys
import copy
import numpy as np
from pysb import Monomer, Parameter, Initial, Observable, Rule, Expression
from pysb.simulator import ScipyOdeSimulator
from pysb.examples import robertson, earm_1_0
import unittest


class TestScipySimulatorBase(object):
    @with_model
    def setUp(self):
        Monomer('A', ['a'])
        Monomer('B', ['b'])

        Parameter('ksynthA', 100)
        Parameter('ksynthB', 100)
        Parameter('kbindAB', 100)

        Parameter('A_init', 0)
        Parameter('B_init', 0)

        Initial(A(a=None), A_init)
        Initial(B(b=None), B_init)

        Observable("A_free", A(a=None))
        Observable("B_free", B(b=None))
        Observable("AB_complex", A(a=1) % B(b=1))

        Rule('A_synth', None >> A(a=None), ksynthA)
        Rule('B_synth', None >> B(b=None), ksynthB)
        Rule('AB_bind', A(a=None) + B(b=None) >> A(a=1) % B(b=1), kbindAB)

        self.model = model

        # Convenience shortcut for accessing model monomer objects
        self.mon = lambda m: self.model.monomers[m]

        # This timespan is chosen to be enough to trigger a Jacobian evaluation
        # on the various solvers.
        self.time = np.linspace(0, 1)
        self.sim = ScipyOdeSimulator(self.model, tspan=self.time,
                                     integrator='vode')

    def tearDown(self):
        self.model = None
        self.time = None
        self.sim = None


class TestScipySimulatorSingle(TestScipySimulatorBase):
    def test_vode_solver_run(self):
        """Test vode."""
        simres = self.sim.run()
        assert simres._nsims == 1

    def test_lsoda_solver_run(self):
        """Test lsoda."""
        solver_lsoda = ScipyOdeSimulator(self.model, tspan=self.time,
                                         integrator='lsoda')
        solver_lsoda.run()

    def test_lsoda_jac_solver_run(self):
        """Test lsoda and analytic jacobian."""
        solver_lsoda_jac = ScipyOdeSimulator(self.model, tspan=self.time,
                                             integrator='lsoda',
                                             use_analytic_jacobian=True)
        solver_lsoda_jac.run()

    def test_y0_as_list(self):
        """Test y0 with list of initial conditions"""
        # Test the initials getter method before anything is changed
        assert np.allclose(self.sim.initials[0][0:2],
                           [ic[1].value for ic in
                            self.model.initial_conditions])

        initials = [10, 20, 0]
        simres = self.sim.run(initials=initials)
        assert np.allclose(simres.initials[0], initials)
        assert np.allclose(simres.observables['A_free'][0], 10)

    def test_y0_as_ndarray(self):
        """Test y0 with numpy ndarray of initial conditions"""
        simres = self.sim.run(initials=np.asarray([10, 20, 0]))
        assert np.allclose(simres.observables['A_free'][0], 10)

    def test_y0_as_dictionary_monomer_species(self):
        """Test y0 with model-defined species."""
        self.sim.initials = {self.mon('A')(a=None): 17}
        base_initials = self.sim.initials
        assert base_initials[0][0] == 17

        simres = self.sim.run(initials={self.mon('A')(a=None): 10,
                               self.mon('B')(b=1) % self.mon('A')(a=1): 0,
                               self.mon('B')(b=None): 0})
        assert np.allclose(simres.initials, [10, 0, 0])
        assert np.allclose(simres.observables['A_free'][0], 10)

        # Initials should reset to base values
        assert np.allclose(self.sim.initials, base_initials)

    def test_y0_as_dictionary_with_bound_species(self):
        """Test y0 with dynamically generated species."""
        simres = self.sim.run(initials={self.mon('A')(a=None): 0,
                               self.mon('B')(b=1) % self.mon('A')(a=1): 100,
                               self.mon('B')(b=None): 0})
        assert np.allclose(simres.observables['AB_complex'][0], 100)

    @raises(TypeError)
    def test_y0_non_numeric_value(self):
        """Test y0 with non-numeric value."""
        self.sim.run(initials={self.mon('A')(a=None): 'eggs'})

    def test_param_values_as_dictionary(self):
        """Test param_values as a dictionary."""
        simres = self.sim.run(param_values={'kbindAB': 0})
        # kbindAB=0 should ensure no AB_complex is produced.
        assert np.allclose(simres.observables["AB_complex"], 0)

    def test_param_values_as_list_ndarray(self):
        """Test param_values as a list and ndarray."""
        orig_param_values = self.sim.param_values
        param_values = [50, 60, 70, 0, 0]
        self.sim.param_values = param_values
        simres = self.sim.run()
        assert np.allclose(self.sim.param_values, param_values)
        assert np.allclose(simres.param_values, param_values)
        # Reset to original param values
        self.sim.param_values = orig_param_values

        # Same thing, but with a numpy array, applied as a run argument
        param_values = np.asarray([55, 65, 75, 0, 0])
        simres = self.sim.run(param_values=param_values)
        assert np.allclose(simres.param_values, param_values)
        # param_values should reset to originals after the run
        assert np.allclose(self.sim.param_values, orig_param_values)

    @raises(IndexError)
    def test_param_values_invalid_dictionary_key(self):
        """Test param_values with invalid parameter name."""
        self.sim.run(param_values={'spam': 150})

    @raises(ValueError, TypeError)
    def test_param_values_non_numeric_value(self):
        """Test param_values with non-numeric value."""
        self.sim.run(param_values={'ksynthA': 'eggs'})

    def test_result_dataframe(self):
        df = self.sim.run().dataframe


class TestScipyOdeCompilerTests(TestScipySimulatorBase):
    """Test vode and analytic jacobian with different compiler backends"""
    def setUp(self):
        super(TestScipyOdeCompilerTests, self).setUp()
        self.args = {'model': self.model,
                     'tspan': self.time,
                     'integrator': 'vode',
                     'use_analytic_jacobian': True}

        self.python_sim = ScipyOdeSimulator(compiler='python', **self.args)
        self.python_res = self.python_sim.run()

    def test_cython(self):
        sim = ScipyOdeSimulator(compiler='cython', **self.args)
        simres = sim.run()
        assert simres.species.shape[0] == self.args['tspan'].shape[0]
        assert np.allclose(self.python_res.dataframe, simres.dataframe)

    def test_theano(self):
        sim = ScipyOdeSimulator(compiler='theano', **self.args)
        simres = sim.run()
        assert simres.species.shape[0] == self.args['tspan'].shape[0]
        assert np.allclose(self.python_res.dataframe, simres.dataframe)

    @unittest.skipIf(sys.version_info.major >= 3, 'weave not available for '
                                                  'Python 3')
    def test_weave(self):
        sim = ScipyOdeSimulator(compiler='weave', **self.args)
        simres = sim.run()
        assert simres.species.shape[0] == self.args['tspan'].shape[0]
        assert np.allclose(self.python_res.dataframe, simres.dataframe)


class TestScipySimulatorSequential(TestScipySimulatorBase):
    def test_sequential_initials(self):
        simres = self.sim.run()
        orig_initials = self.sim.initials

        new_initials = [10, 20, 30]
        simres = self.sim.run(initials=new_initials)

        # Check that single-run initials applied properly to the result
        assert np.allclose(simres.species[0], new_initials)
        assert np.allclose(simres.initials, new_initials)
        # Check that the single-run initials were removed after the run
        assert np.allclose(self.sim.initials, orig_initials)

    def test_sequential_param_values(self):
        orig_param_values = self.sim.param_values
        new_param_values = {'kbindAB': 0}
        new_initials = [15, 25, 35]
        simres = self.sim.run(param_values=new_param_values,
                              initials=new_initials)
        # No new AB_complex should be formed
        assert np.allclose(simres.observables['AB_complex'], new_initials[2])
        assert simres.nsims == 1
        # Original param_values should be restored after run
        assert np.allclose(self.sim.param_values, orig_param_values)

        # Check that per-run param override works when a base param
        # dictionary is also specified
        self.sim.param_values = new_param_values
        base_param_values = new_param_values
        new_param_values = {'ksynthB': 50}
        simres = self.sim.run(param_values=new_param_values)
        # Check that new param value override applied
        assert np.allclose(simres.param_values[0][1],
                           new_param_values['ksynthB'])
        # Check that simulator reverts to base param values
        assert np.allclose(self.sim.param_values[0][2],
                           base_param_values['kbindAB'])
        # Reset to original param values
        self.sim.param_values = orig_param_values

    def test_sequential_tspan(self):
        tspan = np.linspace(0, 10, 11)
        orig_tspan = self.sim.tspan
        simres = self.sim.run(tspan=tspan)
        # Check that new tspan applied properly
        assert np.allclose(simres.tout, tspan)
        # Check that simulator instance reset to original tspan
        assert np.allclose(self.sim.tspan, orig_tspan)


class TestScipySimulatorMultiple(TestScipySimulatorBase):
    def test_initials_and_param_values_two_lists(self):
        initials = [[10, 20, 30], [50, 60, 70]]
        param_values = [[55, 65, 75, 0, 0],
                        [90, 100, 110, 5, 6]]
        import pysb.bng
        pysb.bng.generate_equations(self.sim.model)
        simres = self.sim.run(initials=initials, param_values=param_values)
        assert np.allclose(simres.species[0][0], initials[0])
        assert np.allclose(simres.species[1][0], initials[1])

        assert np.allclose(simres.param_values[0], param_values[0])
        assert np.allclose(simres.param_values[1], param_values[1])

        assert simres.nsims == 2

        # Check the methods underlying these properties work
        df = simres.dataframe
        all = simres.all

        # Try overriding above lists of initials/params with dicts
        self.sim.initials = initials
        self.sim.param_values = param_values
        simres = self.sim.run(
            initials={self.mon('A')(a=None): [103, 104]},
            param_values={'ksynthA': [101, 102]})
        # Simulator initials and params should not persist run() overrides
        assert np.allclose(self.sim.initials, initials)
        assert np.allclose(self.sim.param_values, param_values)
        # Create the expected initials/params arrays and compare to result
        initials = np.array(initials)
        initials[:, 0] = [103, 104]
        param_values = np.array(param_values)
        param_values[:, 0] = [101, 102]
        assert np.allclose(simres.initials, initials)
        assert np.allclose(simres.param_values, param_values)

    @raises(ValueError)
    def test_run_initials_different_length_to_base(self):
        initials = [[10, 20, 30, 40], [50, 60, 70, 80]]
        self.sim.initials = initials
        self.sim.run(initials=initials[0])

    @raises(ValueError)
    def test_run_params_different_length_to_base(self):
        param_values = [[55, 65, 75, 0, 0, 1],
                        [90, 100, 110, 5, 6, 7]]
        self.sim.param_values = param_values
        self.sim.run(param_values=param_values[0])

    def test_param_values_dict(self):
        param_values = {'A_init': [0, 100]}
        initials = {self.model.monomers['B'](b=None): [250, 350]}

        simres = self.sim.run(param_values=param_values)
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 100])

        simres = self.sim.run(param_values={'B_init': [200, 300]})
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 0])
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'B_free'],
                           [200, 300])

        simres = self.sim.run(initials=initials, param_values=param_values)
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 100])
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'B_free'],
                           [250, 350])

    @raises(ValueError)
    def test_initials_and_param_values_differing_lengths(self):
        initials = [[10, 20, 30, 40], [50, 60, 70, 80]]
        param_values = [[55, 65, 75, 0, 0],
                        [90, 100, 110, 5, 6],
                        [90, 100, 110, 5, 6]]
        self.sim.run(initials=initials, param_values=param_values)


@with_model
def test_integrate_with_expression():
    """Ensure a model with Expressions simulates."""

    Monomer('s1')
    Monomer('s9')
    Monomer('s16')
    Monomer('s20')

    # Parameters should be able to contain s(\d+) without error
    Parameter('ks0',2e-5)
    Parameter('ka20', 1e5)

    Initial(s9(), Parameter('s9_0', 10000))

    Observable('s1_obs', s1())
    Observable('s9_obs', s9())
    Observable('s16_obs', s16())
    Observable('s20_obs', s20())

    Expression('keff', (ks0*ka20)/(ka20+s9_obs))

    Rule('R1', None >> s16(), ks0)
    Rule('R2', None >> s20(), ks0)
    Rule('R3', s16() + s20() >> s16() + s1(), keff)

    time = np.linspace(0, 40)
    sim = ScipyOdeSimulator(model, tspan=time)
    simres = sim.run()
    keff_vals = simres.expressions['keff']
    assert len(keff_vals) == len(time)
    assert np.allclose(keff_vals, 1.8181818181818182e-05)


def test_set_initial_to_zero():
    sim = ScipyOdeSimulator(robertson.model, tspan=np.linspace(0, 100))
    simres = sim.run(initials={robertson.model.monomers['A'](): 0})
    assert np.allclose(simres.observables['A_total'], 0)


def test_robertson_integration():
    """Ensure robertson model simulates."""
    t = np.linspace(0, 100)
    # Run with or without inline
    sim = ScipyOdeSimulator(robertson.model)
    simres = sim.run(tspan=t)
    assert simres.species.shape[0] == t.shape[0]
    if sim._compiler != 'python':
        # Also run without inline
        sim = ScipyOdeSimulator(robertson.model, tspan=t, compiler='python')
        simres = sim.run()
        assert simres.species.shape[0] == t.shape[0]


def test_earm_integration():
    """Ensure earm_1_0 model simulates."""
    t = np.linspace(0, 1e3)
    # Run with or without inline
    sim = ScipyOdeSimulator(earm_1_0.model, tspan=t)
    sim.run()
    if sim._compiler != 'python':
        # Also run without inline
        ScipyOdeSimulator(earm_1_0.model, tspan=t, compiler='python').run()


@raises(ValueError)
def test_simulation_no_tspan():
    ScipyOdeSimulator(robertson.model).run()


@raises(UserWarning)
def test_nonexistent_integrator():
    """Ensure nonexistent integrator raises."""
    ScipyOdeSimulator(robertson.model, tspan=np.linspace(0, 1, 2),
                      integrator='does_not_exist')


def test_unicode_obsname_ascii():
    """Ensure ascii-convetible unicode observable names are handled."""
    t = np.linspace(0, 100)
    rob_copy = copy.deepcopy(robertson.model)
    rob_copy.observables[0].name = u'A_total'
    sim = ScipyOdeSimulator(rob_copy)
    simres = sim.run(tspan=t)
    simres.all
    simres.dataframe


if sys.version_info[0] < 3:
    @raises(ValueError)
    def test_unicode_obsname_nonascii():
        """Ensure non-ascii unicode observable names error in python 2."""
        t = np.linspace(0, 100)
        rob_copy = copy.deepcopy(robertson.model)
        rob_copy.observables[0].name = u'A_total\u1234'
        sim = ScipyOdeSimulator(rob_copy)
        simres = sim.run(tspan=t)


def test_unicode_exprname_ascii():
    """Ensure ascii-convetible unicode expression names are handled."""
    t = np.linspace(0, 100)
    rob_copy = copy.deepcopy(robertson.model)
    ab = rob_copy.observables['A_total'] + rob_copy.observables['B_total']
    expr = Expression(u'A_plus_B', ab, _export=False)
    rob_copy.add_component(expr)
    sim = ScipyOdeSimulator(rob_copy)
    simres = sim.run(tspan=t)
    simres.all
    simres.dataframe


if sys.version_info[0] < 3:
    @raises(ValueError)
    def test_unicode_exprname_nonascii():
        """Ensure non-ascii unicode expression names error in python 2."""
        t = np.linspace(0, 100)
        rob_copy = copy.deepcopy(robertson.model)
        ab = rob_copy.observables['A_total'] + rob_copy.observables['B_total']
        expr = Expression(u'A_plus_B\u1234', ab, _export=False)
        rob_copy.add_component(expr)
        sim = ScipyOdeSimulator(rob_copy)
        simres = sim.run(tspan=t)




from pysb.testing import *
from pysb.core import *
from pysb.macros import *

@with_model
def test_catalyze_one_step_None():
    # What if no substrate is required, hand-wavy, but present in published models!
    Monomer('E', ['b'])
    Monomer('P')
    catalyze_one_step(E, None, P, 1e-4)
    assert_equal(len(model.rules), 1)